
"""
Tests for the upgrade of L{PrivateApplication} schema from 4 to 5, which was
dropping the 'hitCount' variable and adding the L{IWebViewer} powerup.
"""

from xmantissa.test.historic.test_privateApplication3to4 import (
    PrivateApplicationUpgradeTests)

class PrivateApplicationUpgradeTests5(PrivateApplicationUpgradeTests):
    """
    Tests for L{xmantissa.webapp.privateApplication4to5}.
    """





from axiom.test.historic import stubloader
from xmantissa.stats import StatBucket

class FreeTicketSignupTestCase(stubloader.StubbedTest):
    def testUpgrade(self):
        for bucket in self.store.query(StatBucket):
            self.assertEqual(bucket.type, "axiom_commits")



# -*- test-case-name: xmantissa.test.historic -*-

import pytz

from zope.interface import implements

from twisted.python.components import registerAdapter

from nevow import athena, tags, loaders, inevow
from nevow.taglibrary import tabbedPane
from nevow.page import renderer

from axiom.item import Item
from axiom import attributes, upgrade

from xmantissa.webtheme import getLoader
from xmantissa.liveform import ChoiceParameter, LiveForm
from xmantissa import webnav, ixmantissa
from xmantissa.webgestalt import AuthenticationApplication

class PreferenceCollectionMixin:
    """
    Convenience mixin for L{xmantissa.ixmantissa.IPreferenceCollection}
    implementors.  Provides only the C{getPreferences} method.
    """

    def getPreferences(self):
        # this is basically a hack so that PreferenceAggregator can
        # continue work in a similar way
        d = {}
        for param in self.getPreferenceParameters() or ():
            d[param.name] = getattr(self, param.name)
        return d

class DefaultPreferenceCollection(Item, PreferenceCollectionMixin):
    """
    Badly named L{xmantissa.ixmantissa.IPreferenceCollection} which
    encapsulates basic preferences that are useful to Mantissa in
    various places, and probably to Mantissa-based applications as well.
    """

    implements(ixmantissa.IPreferenceCollection)

    typeName = 'mantissa_default_preference_collection'
    schemaVersion = 2

    installedOn = attributes.reference()

    itemsPerPage = attributes.integer(default=10)
    timezone = attributes.text(default=u'US/Eastern')

    powerupInterfaces = (ixmantissa.IPreferenceCollection,)

    def getPreferenceParameters(self):
        return (ChoiceParameter(
                    'timezone',
                    list((c, unicode(c, 'ascii'), c == self.timezone) for c
                            in pytz.common_timezones),
                    'Timezone'),
                ChoiceParameter(
                    'itemsPerPage',
                    list((c, c, c == self.itemsPerPage) for c in (10, 20, 30)),
                    'Items Per Page'))

    def getSections(self):
        authapp = self.store.findUnique(AuthenticationApplication, default=None)
        if authapp is None:
            return None
        return (authapp,)

    def getTabs(self):
        return (webnav.Tab('General', self.storeID, 1.0, authoritative=True),)


upgrade.registerAttributeCopyingUpgrader(DefaultPreferenceCollection, 1, 2)


class PreferenceCollectionLiveForm(LiveForm):
    """
    L{xmantissa.liveform.LiveForm} subclass which switches
    the docfactory, the jsClass, and overrides the submit
    button renderer.
    """
    jsClass = u'Mantissa.Preferences.PrefCollectionLiveForm'

    def __init__(self, *a, **k):
        super(PreferenceCollectionLiveForm, self).__init__(*a, **k)
        self.docFactory = getLoader('liveform-compact')

    def submitbutton(self, request, tag):
        return tags.input(type='submit', name='__submit__', value='Save')
    renderer(submitbutton)


class PreferenceCollectionFragment(athena.LiveElement):
    """
    L{inevow.IRenderer} adapter for L{xmantissa.ixmantissa.IPreferenceCollection}.
    """
    docFactory = loaders.stan(tags.directive('fragments'))
    liveFormClass = PreferenceCollectionLiveForm

    def __init__(self, collection):
        super(PreferenceCollectionFragment, self).__init__()
        self.collection = collection

    def fragments(self, req, tag):
        """
        Render our preference collection, any child preference
        collections we discover by looking at self.tab.children,
        and any fragments returned by its C{getSections} method.

        Subtabs and C{getSections} fragments are rendered as fieldsets
        inside the parent preference collection's tab.
        """
        f = self._collectionToLiveform()
        if f is not None:
            yield tags.fieldset[tags.legend[self.tab.name], f]

        for t in self.tab.children:
            f = inevow.IRenderer(
                    self.collection.store.getItemByID(t.storeID))
            f.tab = t
            if hasattr(f, 'setFragmentParent'):
                f.setFragmentParent(self)
            yield f

        for f in self.collection.getSections() or ():
            f = ixmantissa.INavigableFragment(f)
            f.setFragmentParent(self)
            f.docFactory = getLoader(f.fragmentName)
            yield tags.fieldset[tags.legend[f.title], f]
    renderer(fragments)

    def _collectionToLiveform(self):
        params = self.collection.getPreferenceParameters()
        if not params:
            return None
        f = self.liveFormClass(
                lambda **k: self._savePrefs(params, k),
                params,
                description=self.tab.name)
        f.setFragmentParent(self)
        return f

    def _savePrefs(self, params, values):
        for (k, v) in values.iteritems():
            setattr(self.collection, k, v)
        return self._collectionToLiveform()

# IRenderer(IPreferenceCollection)
#   -> PreferenceCollectionFragment, unless an adapter is registered for a
# specific IPreferenceCollection implementor
registerAdapter(PreferenceCollectionFragment,
                ixmantissa.IPreferenceCollection,
                inevow.IRenderer)

class PreferenceAggregator(Item):
    """
    L{xmantissa.ixmantissa.IPreferenceAggregator} implementor,
    which provides access to the values of preferences by name
    """
    implements(ixmantissa.IPreferenceAggregator)

    schemaVersion = 1
    typeName = 'preference_aggregator'

    _collections = attributes.inmemory()
    installedOn = attributes.reference()

    powerupInterfaces = (ixmantissa.IPreferenceAggregator,)

    def activate(self):
        self._collections = None

    # IPreferenceAggregator
    def getPreferenceCollections(self):
        if self._collections is None:
            self._collections = list(self.store.powerupsFor(ixmantissa.IPreferenceCollection))
        return self._collections

    def getPreferenceValue(self, key):
        for collection in self.getPreferenceCollections():
            for (_key, value) in collection.getPreferences().iteritems():
                if _key == key:
                    return value


class PreferenceEditor(athena.LiveElement):
    """
    L{xmantissa.ixmantissa.INavigableFragment} adapter for
    L{xmantissa.prefs.PreferenceAggregator}.  Responsible for
    rendering all installed L{xmantissa.ixmantissa.IPreferenceCollection}s
    """
    implements(ixmantissa.INavigableFragment)
    title = 'Settings'
    fragmentName = 'preference-editor'

    def __init__(self, aggregator):
        self.aggregator = aggregator
        super(PreferenceEditor, self).__init__()

    def tabbedPane(self, req, tag):
        """
        Render a tabbed pane tab for each top-level
        L{xmantissa.ixmantissa.IPreferenceCollection} tab
        """
        navigation = webnav.getTabs(self.aggregator.getPreferenceCollections())
        pages = list()
        for tab in navigation:
            f = inevow.IRenderer(
                    self.aggregator.store.getItemByID(tab.storeID))
            f.tab = tab
            if hasattr(f, 'setFragmentParent'):
                f.setFragmentParent(self)
            pages.append((tab.name, f))

        f = tabbedPane.TabbedPaneFragment(pages, name='preference-editor')
        f.setFragmentParent(self)
        return f
    renderer(tabbedPane)

    def head(self):
        return tabbedPane.tabbedPaneGlue.inlineCSS

registerAdapter(PreferenceEditor, PreferenceAggregator, ixmantissa.INavigableFragment)



# -*- test-case-name: xmantissa.test.historic.test_privateApplication3to4 -*-

"""
Generate a test database containing a L{PrivateApplication} installed on its
store without powering it up for L{ITemplateNameResolver}.
"""

from axiom.test.historic.stubloader import saveStub
from axiom.dependency import installOn
from axiom.userbase import LoginSystem

from xmantissa.webapp import PrivateApplication

USERNAME = u'testuser'
DOMAIN = u'localhost'
PREFERRED_THEME = u'theme-preference'
HIT_COUNT = 8765
PRIVATE_KEY = 123456


def createDatabase(store):
    """
    Instantiate a L{PrivateApplication} in C{store} and install it.
    """
    loginSystem = LoginSystem(store=store)
    installOn(loginSystem, store)
    account = loginSystem.addAccount(USERNAME, DOMAIN, None)
    subStore = account.avatars.open()

    app = PrivateApplication(
        store=subStore,
        preferredTheme=PREFERRED_THEME,
        privateKey=PRIVATE_KEY)
    installOn(app, subStore)


if __name__ == '__main__':
    saveStub(createDatabase, 16534)




from axiom.test.historic.stubloader import StubbedTest
from axiom.batch import processor

from xmantissa.fulltext import HypeIndexer, XapianIndexer, PyLuceneIndexer
from xmantissa.test.historic.stub_remoteIndexer1to2 import StubSource


class RemoteIndexerTestCase(StubbedTest):
    """
    Test that each kind of remote indexer correctly becomes associated with an
    item source when being upgraded to version two.
    """
    def testUpgradeHype(self):
        indexer = self.store.findUnique(HypeIndexer)
        self.assertEquals(
            [self.store.findUnique(StubSource)],
            list(indexer.getSources()))


    def testUpgradeXapian(self):
        indexer = self.store.findUnique(XapianIndexer)
        self.assertEquals(
            [self.store.findUnique(StubSource)],
            list(indexer.getSources()))


    def testUpgradePyLucene(self):
        indexer = self.store.findUnique(PyLuceneIndexer)
        self.assertEquals(
            [self.store.findUnique(StubSource)],
            list(indexer.getSources()))



# -*- test-case-name: xmantissa.test.test_interstore -*-

"""
This module provides an implementation-agnostic mechanism for routing messages
between users.  Mantissa requires such a mechanism for any truly multi-user
application, because Axiom partitions each user's account into its own store.
For more detail on specifics of that separation, see the module documentation
for L{axiom.userbase}.

Let's say we wanted to write an application which allowed Alice
<alice@somewhere.example.com> to make an appointment with Bob
<bob@elsewhere.example.com>.  We'll call it 'Example App'.

    - Write the code for your Calendar item.  It should implement
      L{IMessageReceiver} such that it can receive appointment requests.

    - Create these hypothetical calendar items in both Bob and Alice's stores.
      These should also be shared - specifically its IMessageReceiver interface
      must be shared, at least from Alice to Bob and from Bob to Alice.  Give
      it a shareID qualified with the name of the application - for example,
      "exampleapp.root.calendar" - to avoid conflicts with other applications.

    - Install a L{MessageQueue} powerup for both alice and bob.

    - Now, Alice wants to request a meeting with bob.  She should send a
      message to a target like C{Identifier(u'exampleapp.root.calendar',
      u'bob', u'elsewhere.example.com')}, and of course the sender should be
      C{Identifier(u'exampleapp.root.calendar', u'alice',
      u'somewhere.example.com')}, using her L{MessageQueue} powerup's
      L{queueMessage} method.  Most applications will want to call this
      indirectly, via the high-level interface in
      L{AMPMessenger.messageRemote}.

    - Bob's Calendar item will receive this message.  It parses the contents of
      the message into a structured meeting request and performs the
      appropriate transaction.  You can use AMP formatting and argument parsing
      to format and parse message bodies by using the L{AMPMessenger} and
      L{AMPReceiver} classes provided here, respectively.

    - If Bob's calendar needs to respond to Alice's calendar to confirm, it can
      use the queueMessage function to reply.

This model of inter-user communciation is a bit more work than simply
manipulating a shared database, but the loose coupling it enforces provides a
long list of advantages.  System upgrades, for example, can be performed
incrementally, one account at a time, and newer versions can talk to older
versions using a compatible protocol.  The message routing mechanism is
pluggable, and it is possible to implement versions which talk to multiple
Mantissa processes to take advantage of multiple CPU cores, multiple mounted
disks, different Mantissa deployments across the internet, or multiple Mantissa
nodes of the same application within a cluster.  Although these things don't
come "out of the box", the programming model here allows all of these changes
to be made without changing any of your application code, just the high-level
routing glue.
"""

from datetime import timedelta

from zope.interface import implements

from twisted.python import log
from twisted.python.failure import Failure
from twisted.internet import defer
from twisted.protocols.amp import COMMAND, ERROR, Argument, Box, parseString

from epsilon.structlike import record

from epsilon.expose import Exposer

from axiom.iaxiom import IScheduler
from axiom.item import Item, declareLegacyItem
from axiom.errors import UnsatisfiedRequirement
from axiom.attributes import text, bytes, integer, AND, reference, inmemory
from axiom.dependency import dependsOn, requiresFromSite
from axiom.userbase import LoginSystem, LoginMethod
from axiom.upgrade import registerUpgrader

from xmantissa.ixmantissa import (
    IMessageRouter, IDeliveryConsequence, IMessageReceiver)

from xmantissa.error import (
    MessageTransportError, BadSender, UnknownMessageType, RevertAndRespond,
    MalformedMessage)

from xmantissa.sharing import getPrimaryRole, NoSuchShare, Identifier
from xmantissa._recordattr import RecordAttribute, WithRecordAttributes

DELIVERY_ERROR = u'mantissa.delivery.error'

ERROR_NO_SHARE = 'no-share'
ERROR_NO_USER = 'no-user'
ERROR_REMOTE_EXCEPTION = 'remote-exception'
ERROR_BAD_SENDER = 'bad-sender'

_RETRANSMIT_DELAY = 120



class Value(record('type data')):
    """
    A L{Value} is a combination of a data type and some data of that type.
    It is the content of a message.

    @ivar type: A short string describing the type of the data.  Possible
    values include L{DELIVERY_ERROR}, L{AMP_MESSAGE_TYPE}, or
    L{AMP_ANSWER_TYPE}.
    @type type: L{unicode}

    @ivar data: The payload of the value, parsed according to rules identified
    by C{self.type}.
    @type data: L{str}
    """



class _QueuedMessage(Item, WithRecordAttributes):
    """
    This is a message, queued in the sender's store, awaiting delivery to the
    target.
    """

    senderUsername = text(
        """
        This is the username of the user who is sending the message.
        """,
        allowNone=False)

    senderDomain = text(
        """
        This is the domain name of the user who is sending the message.
        """,
        allowNone=False)

    senderShareID = text(
        """
        This is the shareID of the shared item which is sending the message.
        """)

    sender = RecordAttribute(Identifier,
                             [senderShareID, senderUsername, senderDomain])

    targetUsername = text(
        """
        This is the username of the user which is intended to receive the
        message.
        """,
        allowNone=False)

    targetDomain = text(
        """
        This is the domain name fo the user which is intended to receive the
        message.
        """,
        allowNone=False)

    targetShareID = text(
        """
        This is the target shareID object that the message will be delivered to
        in the foreign store.
        """,
        allowNone=False)

    target = RecordAttribute(Identifier,
                             [targetShareID, targetUsername, targetDomain])

    messageType = text(
        """
        The content-type of the data stored in my L{messageData} attribute.
        """,
        allowNone=False)

    messageData = bytes(
        """
        The data of the message.
        """, allowNone=False)

    value = RecordAttribute(Value,
                               [messageType, messageData])

    messageID = integer(
        """
        An identifier for this message, unique to this store.
        """, allowNone=False)

    consequence = reference(
        """
        A provider of L{IDeliveryConsequence} which will be invoked when the
        answer to this message is received.
        """, allowNone=True)



class _FailedAnswer(Item, WithRecordAttributes):
    """
    A record of an L{answerReceived} method raising an exception.

    There is no way for the system to report the failed processing of an answer
    to its peer (nor should there be), so this class allows a buggy answer
    receiver to remember the answer for later.
    """

    consequence = reference(
        """
        A provider of L{IDeliveryConsequence} which will be invoked when this
        _FailedAnswer is redelivered.
        """,
        allowNone=False)

    messageType = text(
        """
        The content-type of the data stored in my L{messageData} attribute.
        """,
        allowNone=False)

    messageData = bytes(
        """
        The data of the message.
        """, allowNone=False)

    messageValue = RecordAttribute(Value,
                                   [messageType, messageData])
    answerType = text(
        """
        The content-type of the data stored in my L{answerData} attribute.
        """,
        allowNone=False)

    answerData = bytes(
        """
        The data of the answer.
        """, allowNone=False)

    answerValue = RecordAttribute(Value,
                                  [answerType, answerData])

    senderUsername = text(
        """
        This is the username of the user who is sending the message.
        """,
        allowNone=False)

    senderDomain = text(
        """
        This is the domain name of the user who is sending the message.
        """,
        allowNone=False)

    senderShareID = text(
        """
        This is the shareID of the shared item which is sending the message.
        """)

    sender = RecordAttribute(Identifier,
                             [senderShareID, senderUsername, senderDomain])

    targetUsername = text(
        """
        This is the username of the user which is intended to receive the
        message.
        """,
        allowNone=False)

    targetDomain = text(
        """
        This is the domain name fo the user which is intended to receive the
        message.
        """,
        allowNone=False)

    targetShareID = text(
        """
        This is the target shareID object that the message will be delivered to
        in the foreign store.
        """,
        allowNone=False)

    target = RecordAttribute(Identifier,
                             [targetShareID, targetUsername, targetDomain])


    def redeliver(self):
        """
        Re-deliver the answer to the consequence which previously handled it
        by raising an exception.

        This method is intended to be invoked after the code in question has
        been upgraded.  Since there are no buggy answer receivers in
        production, nothing calls it yet.
        """
        self.consequence.answerReceived(self.answerValue,
                                        self.messageValue,
                                        self.sender,
                                        self.target)
        self.deleteFromStore()



class _AlreadyAnswered(Item, WithRecordAttributes):
    """
    An L{AlreadyAnswered} is a persistent record of an answer to a message
    delivered via L{queueMessage}.  This stays in the database until the
    original sender has sent a definitive acknowledgement to this answer, so
    that duplicate L{routeMessage} invocations do not cause duplicate
    application-level processing of the message.

    @ivar deliveryDeferred: a L{Deferred} in memory, representing a currently
    pending delivery attempt.  This is L{None} if no delivery attempt is
    currently pending.
    """

    deliveryDeferred = inmemory()


    def activate(self):
        """
        Initialize in-memory state.
        """
        self.deliveryDeferred = None


    originalSenderShareID = text(
        """
        This is the shareID of the item which originally sent the message that
        this answer is in response to; the one that the answer is being
        delivered to.
        """,
        allowNone=True)

    originalSenderUsername = text(
        """
        This is the localpart of the user's account which originally sent the
        message that this answer is in response to; the one that the answer is
        being delivered to.
        """,
        allowNone=False)

    originalSenderDomain = text(
        """
        This is the domain name of the user's account which originally sent the
        message that this answer is in response to; the one that the answer is
        being delivered to.
        """,
        allowNone=False)


    originalSender = RecordAttribute(Identifier,
                                     [originalSenderShareID,
                                      originalSenderUsername,
                                      originalSenderDomain])

    originalTargetShareID = text(
        """
        This is the shareID of the original item which received the message
        that this answer in response to.  This is the item that the response is
        being sent from.
        """,
        allowNone=False)

    originalTargetUsername = text(
        """
        This is the localpart of the original account which received the
        message that this answer in response to.  This is the user that the
        response is being sent from.
        """,
        allowNone=False)

    originalTargetDomain = text(
        """
        This is the domain name of the original account which received the
        message that this answer in response to.  This is the user that the
        response is being sent from.
        """,
        allowNone=False)

    originalTarget = RecordAttribute(Identifier,
                                     [originalTargetShareID,
                                      originalTargetUsername,
                                      originalTargetDomain])

    messageID = integer(
        """
        An identifier, unique to the original sender account name
        (username@domain) for this message.
        """, allowNone=False)

    answerType = text(
        """
        Some text, identifying the type of the answer.  This refers to the data
        in L{answerData} - this attribute will be set to L{DELIVERY_ERROR} if
        the answer could not be delivered.
        """,
        allowNone=False)

    answerData = bytes(
        """
        The data returned from the original receiver of this answer.
        """,
        allowNone=False)

    value = RecordAttribute(Value,
                               [answerType, answerData])



class _NullRouter(object):
    """
    A null L{IMessageRouter} implementation, which drops messages on the floor.
    This is only used in the case where an L{IMessageRouter} powerup cannot be
    found on the site store to route messages to their recipients.
    """

    implements(IMessageRouter)

    def routeAnswer(self, *a):
        """
        Route an answer, but drop it on the floor.
        """
        return defer.fail(MessageTransportError())


    def routeMessage(self, *a):
        """
        Route a message by dropping it on the floor.
        """


class LocalMessageRouter(record("loginSystem")):
    """
    This is an item installed on the site store to route messages to
    appropriate users.  This only routes messages between different stores that
    are referred to as accounts by a L{LoginSystem} on a single node.  In other
    words it implements only local delivery.
    """

    implements(IMessageRouter)


    def _routerForAccount(self, identifier):
        """
        Locate an avatar by the username and domain portions of an
        L{Identifier}, so that we can deliver a message to the appropriate
        user.
        """
        acct = self.loginSystem.accountByAddress(identifier.localpart,
                                                 identifier.domain)
        return IMessageRouter(acct, None)


    def routeMessage(self, sender, target, value, messageID):
        """
        Implement L{IMessageRouter.routeMessage} by synchronously locating an
        account via L{axiom.userbase.LoginSystem.accountByAddress}, and
        delivering a message to it by calling a method on it.
        """
        router = self._routerForAccount(target)
        if router is not None:
            router.routeMessage(sender, target, value, messageID)
        else:
            reverseRouter = self._routerForAccount(sender)
            reverseRouter.routeAnswer(sender, target,
                                      Value(DELIVERY_ERROR, ERROR_NO_USER),
                                      messageID)


    def routeAnswer(self, originalSender, originalTarget, value, messageID):
        """
        Implement L{IMessageRouter.routeMessage} by synchronously locating an
        account via L{axiom.userbase.LoginSystem.accountByAddress}, and
        delivering a response to it by calling a method on it and returning a
        deferred containing its answer.
        """
        router = self._routerForAccount(originalSender)
        return router.routeAnswer(originalSender, originalTarget,
                                  value, messageID)



def _accidentalSiteRouter(siteStore):
    """
    Create an L{IMessageRouter} provider for an item in a user store
    accidentally opened as a site store.
    """
    try:
        raise UnsatisfiedRequirement()
    except UnsatisfiedRequirement:
        log.err(Failure(),
                "You have opened a user's store as if it were a site store.  "
                "Message routing is disabled.")
    return _NullRouter()



def _createLocalRouter(siteStore):
    """
    Create an L{IMessageRouter} provider for the default case, where no
    L{IMessageRouter} powerup is installed on the top-level store.

    It wraps a L{LocalMessageRouter} around the L{LoginSystem} installed on the
    given site store.

    If no L{LoginSystem} is present, this returns a null router which will
    simply log an error but not deliver the message anywhere, until this
    configuration error can be corrected.

    @rtype: L{IMessageRouter}
    """
    ls = siteStore.findUnique(LoginSystem, default=None)
    if ls is None:
        try:
            raise UnsatisfiedRequirement()
        except UnsatisfiedRequirement:
            log.err(Failure(),
                    "You have opened a substore from a site store with no "
                    "LoginSystem.  Message routing is disabled.")
        return _NullRouter()
    return LocalMessageRouter(ls)



class MessageQueue(Item):
    """
    A queue of outgoing L{QueuedMessage} objects.
    """
    schemaVersion = 2
    powerupInterfaces = (IMessageRouter,)

    siteRouter = requiresFromSite(IMessageRouter,
                                  _createLocalRouter,
                                  _accidentalSiteRouter)

    messageCounter = integer(
        """
        This counter generates identifiers for outgoing messages.
        """,
        default=0, allowNone=False)


    def _scheduleMePlease(self):
        """
        This queue needs to have its run() method invoked at some point in the
        future.  Tell the dependent scheduler to schedule it if it isn't
        already pending execution.
        """
        sched = IScheduler(self.store)
        if len(list(sched.scheduledTimes(self))) == 0:
            sched.schedule(self, sched.now())


    def routeMessage(self, sender, target, value, messageID):
        """
        Implement L{IMessageRouter.routeMessage} by locating a shared item
        which provides L{IMessageReceiver}, identified by L{target} in this
        L{MessageQueue}'s L{Store}, as shared to the specified C{sender}, then
        invoke its L{messageReceived} method.  Then, take the results of that
        L{messageReceived} invocation and deliver them as an answer to the
        object specified by L{sender}.

        If any of these steps fail such that no
        L{IMessageReceiver.messageReceived} method may be invoked, generate a
        L{DELIVERY_ERROR} response instead.
        """
        avatarName = sender.localpart + u"@" + sender.domain
        # Look for the sender.
        answer = self.store.findUnique(
            _AlreadyAnswered,
            AND(_AlreadyAnswered.originalSender == sender,
                _AlreadyAnswered.messageID == messageID),
            default=None)
        if answer is None:
            role = getPrimaryRole(self.store, avatarName)
            try:
                receiver = role.getShare(target.shareID)
            except NoSuchShare:
                response = Value(DELIVERY_ERROR,  ERROR_NO_SHARE)
            else:
                try:
                    def txn():
                        output = receiver.messageReceived(value, sender,
                                                          target)
                        if not isinstance(output, Value):
                            raise TypeError("%r returned non-Value %r" %
                                            (receiver, output))
                        return output
                    response = self.store.transact(txn)
                except RevertAndRespond, rar:
                    response = rar.value
                except:
                    log.err(Failure(),
                            "An error occurred during inter-store "
                            "message delivery.")
                    response = Value(DELIVERY_ERROR, ERROR_REMOTE_EXCEPTION)
            answer = _AlreadyAnswered.create(store=self.store,
                                             originalSender=sender,
                                             originalTarget=target,
                                             messageID=messageID,
                                             value=response)
        self._deliverAnswer(answer)
        self._scheduleMePlease()


    def _deliverAnswer(self, answer):
        """
        Attempt to deliver an answer to a message sent to this store, via my
        store's parent's L{IMessageRouter} powerup.

        @param answer: an L{AlreadyAnswered} that contains an answer to a
        message sent to this store.
        """
        router = self.siteRouter
        if answer.deliveryDeferred is None:
            d = answer.deliveryDeferred = router.routeAnswer(
                answer.originalSender, answer.originalTarget, answer.value,
                answer.messageID)
            def destroyAnswer(result):
                answer.deleteFromStore()
            def transportErrorCheck(f):
                answer.deliveryDeferred = None
                f.trap(MessageTransportError)
            d.addCallbacks(destroyAnswer, transportErrorCheck)
            d.addErrback(log.err)


    def routeAnswer(self, originalSender, originalTarget, value, messageID):
        """
        Route an incoming answer to a message originally sent by this queue.
        """
        def txn():
            qm = self._messageFromSender(originalSender, messageID)
            if qm is None:
                return
            c = qm.consequence
            if c is not None:
                c.answerReceived(value, qm.value,
                                 qm.sender, qm.target)
            elif value.type == DELIVERY_ERROR:
                try:
                    raise MessageTransportError(value.data)
                except MessageTransportError:
                    log.err(Failure(),
                            "An unhandled delivery error occurred on a message"
                            " with no consequence.")
            qm.deleteFromStore()
        try:
            self.store.transact(txn)
        except:
            log.err(Failure(),
                    "An unhandled error occurred while handling a response to "
                    "an inter-store message.")
            def answerProcessingFailure():
                qm = self._messageFromSender(originalSender, messageID)
                _FailedAnswer.create(store=qm.store,
                                     consequence=qm.consequence,
                                     sender=originalSender,
                                     target=originalTarget,
                                     messageValue=qm.value,
                                     answerValue=value)
                qm.deleteFromStore()
            self.store.transact(answerProcessingFailure)
        return defer.succeed(None)


    def _messageFromSender(self, sender, messageID):
        """
        Locate a previously queued message by a given sender and messageID.
        """
        return self.store.findUnique(
            _QueuedMessage,
            AND(_QueuedMessage.senderUsername == sender.localpart,
                _QueuedMessage.senderDomain == sender.domain,
                _QueuedMessage.messageID == messageID),
            default=None)


    def _verifySender(self, sender):
        """
        Verify that this sender is valid.
        """
        if self.store.findFirst(
            LoginMethod,
            AND(LoginMethod.localpart == sender.localpart,
                LoginMethod.domain == sender.domain,
                LoginMethod.internal == True)) is None:
            raise BadSender(sender.localpart + u'@' + sender.domain,
                            [lm.localpart + u'@' + lm.domain
                             for lm in self.store.query(
                        LoginMethod, LoginMethod.internal == True)])


    def queueMessage(self, sender, target, value,
                     consequence=None):
        """
        Queue a persistent outgoing message.

        @param sender: The a description of the shared item that is the sender
        of the message.
        @type sender: L{xmantissa.sharing.Identifier}

        @param target: The a description of the shared item that is the target
        of the message.
        @type target: L{xmantissa.sharing.Identifier}

        @param consequence: an item stored in the same database as this
        L{MessageQueue} implementing L{IDeliveryConsequence}.
        """
        self.messageCounter += 1
        _QueuedMessage.create(store=self.store,
                              sender=sender,
                              target=target,
                              value=value,
                              messageID=self.messageCounter,
                              consequence=consequence)
        self._scheduleMePlease()


    def run(self):
        """
        Attmept to deliver the first outgoing L{QueuedMessage}; return a time
        to reschedule if there are still more retries or outgoing messages to
        send.
        """
        delay = None
        router = self.siteRouter
        for qmsg in self.store.query(_QueuedMessage,
                                     sort=_QueuedMessage.storeID.ascending):
            try:
                self._verifySender(qmsg.sender)
            except:
                self.routeAnswer(qmsg.sender, qmsg.target,
                                 Value(DELIVERY_ERROR, ERROR_BAD_SENDER),
                                 qmsg.messageID)
                log.err(Failure(),
                        "Could not verify sender for sending message.")
            else:
                router.routeMessage(qmsg.sender, qmsg.target,
                                    qmsg.value, qmsg.messageID)

        for answer in self.store.query(_AlreadyAnswered,
                                       sort=_AlreadyAnswered.storeID.ascending):
            self._deliverAnswer(answer)
        nextmsg = self.store.findFirst(_QueuedMessage, default=None)
        if nextmsg is not None:
            delay = _RETRANSMIT_DELAY
        else:
            nextanswer = self.store.findFirst(_AlreadyAnswered, default=None)
            if nextanswer is not None:
                delay = _RETRANSMIT_DELAY
        if delay is not None:
            return IScheduler(self.store).now() + timedelta(seconds=delay)


declareLegacyItem(
    MessageQueue.typeName, 1,
    dict(messageCounter=integer(default=0, allowNone=False),
         scheduler=reference()))

def upgradeMessageQueue1to2(old):
    """
    Copy the C{messageCounter} attribute to the upgraded MessageQueue.
    """
    return old.upgradeVersion(
        MessageQueue.typeName, 1, 2, messageCounter=old.messageCounter)

registerUpgrader(upgradeMessageQueue1to2, MessageQueue.typeName, 1, 2)


#### High-level convenience API ####

AMP_MESSAGE_TYPE = u'mantissa.amp.message'
AMP_ANSWER_TYPE = u'mantissa.amp.answer'


class _ProtoAttributeArgument(Argument):
    """
    Common factoring of L{TargetArgument} and L{SenderArgument}, for reading an
    attribute from the 'proto' attribute.

    @ivar attr: the name of the attribute to retrieve from the C{proto}
    argument to L{_ProtoAttributeArgument.fromBox}.
    """

    def fromBox(self, name, strings, objects, proto):
        """
        Retreive an attribute from the C{proto} parameter.
        """
        objects[name] = getattr(proto, self.attr)


    def toBox(self, name, strings, objects, proto):
        """
        Do nothing; these argument types are for specifying out-of-band
        information not in the message body, so leave the message body alone
        when sending.
        """


class TargetArgument(_ProtoAttributeArgument):
    """
    An AMP L{Argument} which places an L{Identifier} for the target (for
    commands) or original target (for answers) of the message being processed
    into the argument list.
    """
    attr = "target"



class SenderArgument(_ProtoAttributeArgument):
    """
    An AMP L{Argument} which places an L{Identifier} for the sender (for
    commands) or original sender (for answers) of the message being processed
    into the argument list.
    """
    attr = "sender"



class _ProtocolPlaceholder(record("sender target")):
    """
    This placeholder object is passed as the 'proto' object to AMP parsing
    methods, and has the two out-of-band attributes required to support
    L{SenderArgument} and L{TargetArgument}, which are about all you can use it
    for.
    """



class AMPMessenger(record("queue sender target")):
    """
    An L{AMPMessenger} is a conduit between an object sending a message
    (identified by the C{queue} and C{sender} arguments) and a recipient of
    that message (represented by the C{target}) argument.

    @ivar queue: a L{MessageQueue} that will be used to queue messages.

    @ivar sender: an L{Identifier} that will be used as the sender of the
    messages.

    @ivar target: an L{Identifier} that will be used as the target of messages.
    """

    def messageRemote(self, cmdObj, consequence=None, **args):
        """
        Send a message to the peer identified by the target, via the
        given L{Command} object and arguments.

        @param cmdObj: a L{twisted.protocols.amp.Command}, whose serialized
        form will be the message.

        @param consequence: an L{IDeliveryConsequence} provider which will
        handle the result of this message (or None, if no response processing
        is desired).

        @param args: keyword arguments which match the C{cmdObj}'s arguments
        list.

        @return: L{None}
        """
        messageBox = cmdObj.makeArguments(args, self)
        messageBox[COMMAND] = cmdObj.commandName
        messageData = messageBox.serialize()
        self.queue.queueMessage(self.sender, self.target,
                                Value(AMP_MESSAGE_TYPE, messageData),
                                consequence)



class _AMPExposer(Exposer):
    """
    An L{Exposer} whose purpose is to expose objects via L{Command} objects.
    """

    def expose(self, commandObject):
        """
        Declare a method as being related to the given command object.

        @param commandObject: a L{Command} subclass.
        """
        thunk = super(_AMPExposer, self).expose(commandObject.commandName)
        def thunkplus(function):
            result = thunk(function)
            result.command = commandObject
            return result
        return thunkplus


    def responderForName(self, instance, commandName):
        """
        When resolving a command to a method from the wire, the information
        available is the command's name; look up a command.

        @param instance: an instance of a class who has methods exposed via
        this exposer's L{_AMPExposer.expose} method.

        @param commandName: the C{commandName} attribute of a L{Command}
        exposed on the given instance.

        @return: a bound method with a C{command} attribute.
        """
        method = super(_AMPExposer, self).get(instance, commandName)
        return method



class _AMPErrorExposer(Exposer):
    """
    An L{Exposer} whose purpose is to expose objects via L{Command} objects and
    error identifiers.
    """

    def expose(self, commandObject, exceptionType):
        """
        Expose a function for processing a given AMP error.
        """
        thunk = super(_AMPErrorExposer, self).expose(
            (commandObject.commandName,
             commandObject.errors.get(exceptionType)))
        def thunkplus(function):
            result = thunk(function)
            result.command = commandObject
            result.exception = exceptionType
            return result
        return thunkplus


    def errbackForName(self, instance, commandName, errorName):
        """
        Retrieve an errback - a callable object that accepts a L{Failure} as an
        argument - that is exposed on the given instance, given an AMP
        commandName and a name in that command's error mapping.
        """
        return super(_AMPErrorExposer, self).get(instance, (commandName, errorName))



commandMethod = _AMPExposer("""
Use this exposer to expose methods on L{AMPReceiver} subclasses which can
respond to AMP commands.  Use like so::

    @commandMethod.expose(YourCommand)
    def yourMethod(self, yourCommandArgument, ...):
        ...
""")

answerMethod = _AMPExposer("""
Use this exposer to expose methods on L{AMPReceiver} subclasses which can
deal with AMP command responses.  Use like so::

    @answerMethod.expose(YourCommand)
    def yourMethod(self, yourCommandResponseArgument, ...):
        ...
""")

errorMethod = _AMPErrorExposer("""
Use this exposer to expose methods on L{AMPReceiver} subclasses which can
deal with AMP command error responses.  Use like so::

    @errorMethod.expose(YourCommand, YourCommandException)
    def yourMethod(self, failure):
        ...
""")



class AMPReceiver(object):
    """
    This is a mixin for L{Item} objects which wish to implement
    L{IMessageReceiver} and/or L{IDeliveryConsequence} by parsing the bodies of
    arriving messages and answers as AMP boxes.

    To implement L{IMessageReceiver}, use the L{commandMethod} decorator.

    To implement L{IMessageRouter}, use the L{answerMethod} and L{errorMethod}
    decorators.

    For example::

        class MyCommand(Command):
            arguments = [('hello', Integer())]
            response = [('goodbye', Text())]
        class MyResponder(Item, AMPReceiver):
            ...
            @commandMethod.expose(MyCommand)
            def processCommand(self, hello):
                return dict(goodbye=u'goodbye!')
            @answerMethod.expose(MyCommand)
            def processAnswer(self, goodbye):
                # process 'goodbye' here.

    In this example, a L{MyResponder} object might be shared and used as a
    target for a call to L{AMPMessenger.messageRemote} with L{MyCommand}, or
    used as the L{consequence} argument to that same call.
    """
    implements(IMessageReceiver, IDeliveryConsequence)

    def _boxFromData(self, messageData):
        """
        A box.

        @param messageData: a serialized AMP box representing either a message
        or an error.
        @type messageData: L{str}

        @raise MalformedMessage: if the C{messageData} parameter does not parse
        to exactly one AMP box.
        """
        inputBoxes = parseString(messageData)
        if not len(inputBoxes) == 1:
            raise MalformedMessage()
        [inputBox] = inputBoxes
        return inputBox


    def messageReceived(self, value, sender, target):
        """
        An AMP-formatted message was received.  Dispatch to the appropriate
        command responder, i.e. a method on this object exposed with
        L{commandMethod.expose}.

        @see IMessageReceiver.messageReceived
        """
        if value.type != AMP_MESSAGE_TYPE:
            raise UnknownMessageType()
        inputBox = self._boxFromData(value.data)
        thunk = commandMethod.responderForName(self, inputBox[COMMAND])
        placeholder = _ProtocolPlaceholder(sender, target)
        arguments = thunk.command.parseArguments(inputBox, placeholder)
        try:
            result = thunk(**arguments)
        except tuple(thunk.command.errors.keys()), knownError:
            errorCode = thunk.command.errors[knownError.__class__]
            raise RevertAndRespond(
                Value(AMP_ANSWER_TYPE,
                         Box(_error_code=errorCode,
                             _error_description=str(knownError)).serialize()))
        else:
            response = thunk.command.makeResponse(result, None)
            return Value(AMP_ANSWER_TYPE, response.serialize())


    def answerReceived(self, value, originalValue,
                       originalSender, originalTarget):
        """
        An answer was received.  Dispatch to the appropriate answer responder,
        i.e. a method on this object exposed with L{answerMethod.expose}.

        @see IDeliveryConsequence.answerReceived
        """
        if value.type != AMP_ANSWER_TYPE:
            raise UnknownMessageType()
        commandName = self._boxFromData(originalValue.data)[COMMAND]
        rawArgs = self._boxFromData(value.data)
        placeholder = _ProtocolPlaceholder(originalSender, originalTarget)
        if ERROR in rawArgs:
            thunk = errorMethod.errbackForName(self, commandName, rawArgs[ERROR])
            thunk(Failure(thunk.exception()))
        else:
            thunk = answerMethod.responderForName(self, commandName)
            arguments = thunk.command.parseResponse(rawArgs, placeholder)
            thunk(**arguments)


__all__ = ['AMPMessenger', 'AMPReceiver', 'AMP_ANSWER_TYPE', 'AMP_MESSAGE_TYPE',
           'DELIVERY_ERROR', 'ERROR', 'ERROR_BAD_SENDER', 'ERROR_NO_SHARE',
           'ERROR_NO_USER', 'ERROR_REMOTE_EXCEPTION', 'LocalMessageRouter',
           'MessageQueue', 'SenderArgument', 'TargetArgument', 'Value',
           'answerMethod', 'commandMethod', 'errorMethod']




"""
Tests for the upgrade from schema version 1 to 2 of the Person item which
adds the I{vip} attribute.
"""

from xmantissa.test.historic.stub_person1to2 import NAME, CREATED

from axiom.test.historic.stubloader import StubbedTest

from xmantissa.people import Organizer, Person


class PersonUpgradeTests(StubbedTest):
    def test_attributes(self):
        """
        Existing attributes from the L{Person} should still be present and the
        new C{vip} attribute should be C{False}.
        """
        organizer = self.store.findUnique(Organizer)
        person = self.store.findUnique(
            Person, Person.storeID != organizer.storeOwnerPerson.storeID)
        self.assertIdentical(person.organizer, organizer)
        self.assertEqual(person.name, NAME)
        self.assertEqual(person.created, CREATED)
        self.assertEqual(person.vip, False)



# Copyright (c) 2008 Divmod.  See LICENSE for details.

"""
Tests for L{xmantissa.ampserver}.
"""

from zope.interface import Interface, implements
from zope.interface.verify import verifyObject

from twisted.python.failure import Failure
from twisted.internet.defer import Deferred
from twisted.internet.protocol import ServerFactory
from twisted.internet.task import Clock
from twisted.internet import reactor
from twisted.cred.credentials import UsernamePassword
from twisted.protocols.amp import ASK, COMMAND, IBoxReceiver
from twisted.trial.unittest import TestCase

from epsilon.ampauth import CredReceiver
from epsilon.amprouter import _ROUTE
from epsilon.test.test_amprouter import SomeReceiver, CollectingSender
from epsilon.iepsilon import IOneTimePad

from axiom.item import Item
from axiom.store import Store
from axiom.attributes import text, inmemory
from axiom.dependency import installOn
from axiom.userbase import LoginSystem

from xmantissa.ixmantissa import IProtocolFactoryFactory, IBoxReceiverFactory
from xmantissa.ampserver import (
    _RouteConnector, AMPConfiguration,
    AMPAvatar, ProtocolUnknown, Router, Connect, connectRoute,
    EchoFactory, EchoReceiver)

__metaclass__ = type


class AMPConfigurationTests(TestCase):
    """
    Tests for L{xmantissa.ampserver.AMPConfiguration} which defines how to
    create an L{AMP} server.
    """
    def setUp(self):
        """
        Create an in-memory L{Store} with an L{AMPConfiguration} in it.
        """
        self.store = Store()
        self.conf = AMPConfiguration(store=self.store)
        installOn(self.conf, self.store)


    def test_interfaces(self):
        """
        L{AMPConfiguration} implements L{IProtocolFactoryFactory}.
        """
        self.assertTrue(verifyObject(IProtocolFactoryFactory, self.conf))


    def test_powerup(self):
        """
        L{ionstallOn} powers up the target for L{IProtocolFactoryFactory} with
        L{AMPConfiguration}.
        """
        self.assertIn(
            self.conf, list(self.store.powerupsFor(IProtocolFactoryFactory)))


    def test_getFactory(self):
        """
        L{AMPConfiguration.getFactory} returns a L{ServerFactory} instance
        which returns L{CredReceiver} instances from its C{buildProtocol}
        method.
        """
        factory = self.conf.getFactory()
        self.assertTrue(isinstance(factory, ServerFactory))
        protocol = factory.buildProtocol(None)
        self.assertTrue(isinstance(protocol, CredReceiver))


    def test_generateOneTimePad(self):
        """
        L{AMPConfiguration.generateOneTimePad} returns a one-time pad.
        """
        object.__setattr__(self.conf, 'callLater', lambda x, y: None)
        pad = self.conf.generateOneTimePad(self.store)
        self.assertNotEqual(
            pad, self.conf.generateOneTimePad(self.store))


    def test_oneTimePadExpires(self):
        """
        L{AMPConfiguration.generateOneTimePad} should expire its pad.
        """
        def callLater(seconds, f):
            self.assertEqual(
                seconds, self.conf.ONE_TIME_PAD_DURATION)
            f()
        object.__setattr__(self.conf, 'callLater', callLater)
        pad = self.conf.generateOneTimePad(self.store)
        self.assertFalse(pad in self.conf._oneTimePads)



class AMPConfigurationSubStoreTests(TestCase):
    """
    Tests for L{AMPConfiguration} which require a substore.
    """
    def setUp(self):
        """
        Create an in-memory L{Store} with an L{AMPConfiguration} in it, and a
        substore.
        """
        self.store = Store()
        self.conf = AMPConfiguration(store=self.store)
        installOn(self.conf, self.store)

        self.localpart = u'alice'
        self.domain = u'example.org'
        self.password = u'foobar'

        loginSystem = self.store.findUnique(LoginSystem)
        account = loginSystem.addAccount(
            self.localpart, self.domain, self.password, internal=True)
        self.subStore = account.avatars.open()


    def _testPortalLogin(self, credentials):
        factory = self.conf.getFactory()
        protocol = factory.buildProtocol(None)
        portal = protocol.portal

        class IDummy(Interface):
            pass

        avatar = object()
        self.subStore.inMemoryPowerUp(avatar, IDummy)
        login = portal.login(credentials, None, IDummy)
        def cbLoggedIn(result):
            self.assertIdentical(IDummy, result[0])
            self.assertIdentical(avatar, result[1])
        login.addCallback(cbLoggedIn)
        return login


    def test_portal(self):
        """
        L{AMPConfiguration.getFactory} returns a factory which creates
        protocols which have a C{portal} attribute which is a L{Portal} which
        authenticates and authorizes using L{axiom.userbase}.
        """
        return self._testPortalLogin(
            UsernamePassword(
                '%s@%s' % (self.localpart.encode('ascii'),
                           self.domain.encode('ascii')),
                self.password),)


    def test_portalOneTimePad(self):
        """
        L{AMPConfiguration.getFactory} returns a factory which creates
        protocols which have a C{portal} attribute which is a L{Portal} which
        can authenticate using one-time pads.
        """
        object.__setattr__(self.conf, 'callLater', lambda x, y: None)
        PAD = self.conf.generateOneTimePad(self.subStore)
        class OTP:
            implements(IOneTimePad)
            padValue = PAD

        return self._testPortalLogin(OTP())



class StubBoxReceiverFactory(Item):
    """
    L{IBoxReceiverFactory}
    """
    protocol = text()

    receivers = inmemory()
    receiverFactory = SomeReceiver

    def activate(self):
        self.receivers = []


    def getBoxReceiver(self):
        receiver = self.receiverFactory()
        self.receivers.append(receiver)
        return receiver



class AMPAvatarTests(TestCase):
    """
    Tests for L{AMPAvatar} which provides an L{IBoxReceiver} implementation
    that supports routing messages to other L{IBoxReceiver} implementations.
    """
    def setUp(self):
        """
        Create a L{Store} with an L{AMPAvatar} installed on it.
        """
        self.store = Store()
        self.avatar = AMPAvatar(store=self.store)
        installOn(self.avatar, self.store)
        self.factory = StubBoxReceiverFactory(
            store=self.store, protocol=u"bar")
        self.store.powerUp(self.factory, IBoxReceiverFactory)


    def test_interface(self):
        """
        L{AMPAvatar} powers up the item on which it is installed for
        L{IBoxReceiver} and indirects that interface to the real router
        implementation of L{IBoxReceiver}.
        """
        router = IBoxReceiver(self.store)
        self.assertTrue(verifyObject(IBoxReceiver, router))
        self.assertNotIdentical(self.avatar, router)


    def test_connectorStarted(self):
        """
        L{AMPAvatar.indirect} returns a L{Router} with a started route
        connector as its default receiver.
        """
        receiver = SomeReceiver()
        self.avatar.__dict__['connectorFactory'] = lambda router: receiver
        router = self.avatar.indirect(IBoxReceiver)
        router.startReceivingBoxes(object())
        self.assertTrue(receiver.started)


    def test_reactor(self):
        """
        L{AMPAvatar.connectorFactory} returns a L{_RouteConnector} constructed
        using the global reactor.
        """
        connector = self.avatar.connectorFactory(object())
        self.assertIdentical(connector.reactor, reactor)



class RouteConnectorTests(TestCase):
    """
    Tests for L{_RouteConnector}.
    """
    def setUp(self):
        """
        Create a L{Store} with an L{AMPAvatar} installed on it.
        """
        self.clock = Clock()
        self.store = Store()
        self.factory = StubBoxReceiverFactory(
            store=self.store, protocol=u"bar")
        self.store.powerUp(self.factory, IBoxReceiverFactory)
        self.router = Router()
        self.sender = CollectingSender()
        self.connector = _RouteConnector(self.clock, self.store, self.router)
        self.router.startReceivingBoxes(self.sender)
        self.router.bindRoute(self.connector, None).connectTo(None)


    def test_accept(self):
        """
        L{_RouteConnector.accept} returns a C{dict} with a C{'route'} key
        associated with a new route identifier which may be used to send AMP
        boxes to a new instance of the L{IBoxReceiver} indicated by the
        C{protocol} argument passed to C{connect}.
        """
        firstIdentifier = self.connector.accept(
            "first origin", u"bar")['route']
        firstReceiver = self.factory.receivers.pop()
        secondIdentifier = self.connector.accept(
            "second origin", u"bar")['route']
        secondReceiver = self.factory.receivers.pop()
        self.clock.advance(0)

        self.router.ampBoxReceived(
            {_ROUTE: firstIdentifier, 'foo': 'bar'})
        self.router.ampBoxReceived(
            {_ROUTE: secondIdentifier, 'baz': 'quux'})

        self.assertEqual(firstReceiver.boxes, [{'foo': 'bar'}])
        self.assertEqual(secondReceiver.boxes, [{'baz': 'quux'}])


    def test_acceptResponseBeforeApplicationBox(self):
        """
        If the protocol L{_RouteConnector.accept} binds to a new route sends a
        box in its C{startReceivingBoxes} method, that box is sent to the
        network after the I{Connect} response is sent.
        """
        earlyBox = {'foo': 'bar'}

        class EarlyReceiver:
            def startReceivingBoxes(self, sender):
                sender.sendBox(earlyBox)

        object.__setattr__(self.factory, 'receiverFactory', EarlyReceiver)
        self.connector.ampBoxReceived({
                COMMAND: Connect.commandName,
                ASK: 'unique-identifier',
                'origin': 'an origin',
                'protocol': 'bar'})
        self.clock.advance(0)
        self.assertEqual(len(self.sender.boxes), 2)
        route, app = self.sender.boxes
        expectedBox = earlyBox.copy()
        expectedBox[_ROUTE] = 'an origin'
        self.assertEqual(app, expectedBox)


    def test_unknownProtocol(self):
        """
        L{_RouteConnector.accept} raises L{ProtocolUnknown} if passed the name
        of a protocol for which no factory can be found.
        """
        self.assertRaises(
            ProtocolUnknown, self.connector.accept, "origin", u"foo")


    def test_originRoute(self):
        """
        The L{IBoxReceiver}s created by L{_RouteConnector.accept} are started with
        L{IBoxSender}s which are associated with the origin route specified to
        C{accept}.
        """
        origin = u'origin route'
        self.connector.accept(origin, u'bar')
        self.clock.advance(0)

        [bar] = self.factory.receivers
        self.assertTrue(bar.started)
        bar.sender.sendBox({'foo': 'bar'})
        self.assertEqual(self.sender.boxes, [{_ROUTE: origin, 'foo': 'bar'}])

        bar.sender.unhandledError(Failure(RuntimeError("test failure")))
        self.assertEqual(self.sender.errors, ["test failure"])



class ConnectRouteTests(TestCase):
    """
    Tests for L{connectRoute} which implements Mantissa-specific route creation
    logic.
    """
    def test_connectRoute(self):
        """
        L{connectRoute} takes an L{AMP}, a L{Router}, an L{IBoxReceiver} and a
        protocol name and issues a L{Connect} command for that protocol and for
        a newly created route associated with the given receiver over the
        L{AMP}.
        """
        commands = []
        results = []
        class FakeAMP:
            def callRemote(self, cmd, **kw):
                commands.append((cmd, kw))
                results.append(Deferred())
                return results[-1]

        amp = FakeAMP()
        sender = CollectingSender()
        router = Router()
        router.startReceivingBoxes(sender)

        receiver = SomeReceiver()
        protocol = u"proto name"

        d = connectRoute(amp, router, receiver, protocol)

        self.assertEqual(
            commands, [(Connect, {'origin': u'0', 'protocol': u'proto name'})])
        results[0].callback({'route': u'remote route'})

        def cbConnected(receiverAgain):
            self.assertIdentical(receiver, receiverAgain)
            self.assertTrue(receiver.started)

            receiver.sender.sendBox({'foo': 'bar'})
            self.assertEqual(
                sender.boxes, [{_ROUTE: 'remote route', 'foo': 'bar'}])
            router.ampBoxReceived({_ROUTE: '0', 'baz': 'quux'})
            self.assertEqual(receiver.boxes, [{'baz': 'quux'}])
        d.addCallback(cbConnected)
        return d



class BoxReceiverFactoryPowerupTestMixin:
    """
    Common tests for implementors of L{IBoxReceiverFactory}.

    @ivar factoryClass: the L{IBoxReceiverFactory} implementor.
    @ivar protocolClass: An L{IBoxReceiver} implementor.
    """

    def test_factoryInterfaces(self):
        """
        C{self.factoryClass} instances provide L{IBoxReceiverFactory}.
        """
        self.assertTrue(verifyObject(IBoxReceiverFactory, self.factoryClass()))


    def test_factoryPowerup(self):
        """
        When installed, C{self.factoryClass} is a powerup for L{IBoxReceiverFactory}.
        """
        store = Store()
        factory = self.factoryClass(store=store)
        installOn(factory, store)
        self.assertEqual(
            list(store.powerupsFor(IBoxReceiverFactory)),
            [factory])


    def test_getBoxReceiver(self):
        """
        C{self.factoryClass.getBoxReceiver} returns an instance of C{self.protocolClass}.
        """
        receiver = self.factoryClass().getBoxReceiver()
        self.assertTrue(isinstance(receiver, self.protocolClass))


    def test_receiverInterfaces(self):
        """
        C{self.protocolClass} instances provide L{IBoxReceiver}.
        """
        self.assertTrue(verifyObject(IBoxReceiver, self.protocolClass()))



class EchoTests(BoxReceiverFactoryPowerupTestMixin, TestCase):
    """
    Tests for L{EchoFactory} and L{EchoReceiver}, classes which provide a
    simple AMP echo protocol for a Mantissa AMP server.
    """

    factoryClass = EchoFactory
    protocolClass = EchoReceiver

    def test_ampBoxReceived(self):
        """
        L{EchoReceiver.ampBoxReceived} sends the received box back to the
        sender.
        """
        sender = CollectingSender()
        receiver = EchoReceiver()
        receiver.startReceivingBoxes(sender)
        receiver.ampBoxReceived({'foo': 'bar'})
        self.assertEqual(sender.boxes, [{'foo': 'bar'}])
        receiver.stopReceivingBoxes(Failure(Exception("test exception")))



# -*- test-case-name: xmantissa.test.test_rendertools -*-

"""
Simple Nevow-related rendering helpers for use in view tests only.
"""

from nevow.rend import Page
from nevow.athena import LivePage
from nevow.loaders import stan
from nevow.testutil import FakeRequest
from nevow.context import WovenContext
from nevow.inevow import IRequest


class TagTestingMixin:
    """
    Mixin defining various useful tag-related testing functionality.
    """
    def assertTag(self, tag, name, attributes, children):
        """
        Assert that the given tag has the given name, attributes, and children.
        """
        self.assertEqual(tag.tagName, name)
        self.assertEqual(tag.attributes, attributes)
        self.assertEqual(tag.children, children)



def _makeContext():
    """
    Create the request and context objects necessary for rendering a page.

    @return: A two-tuple of the created L{FakeRequest} and L{WovenContext},
    with the former remembered in the latter.
    """
    request = FakeRequest()
    context = WovenContext()
    context.remember(request, IRequest)
    return (request, context)



def renderLiveFragment(fragment):
    """
    Render the given fragment in a LivePage.

    This can only work for fragments which can be rendered synchronously.
    Fragments which involve Deferreds will be silently rendered incompletely.

    @type fragment: L{nevow.athena.LiveFragment} or L{nevow.athena.LiveElement}
    @param fragment: The page component to render.

    @rtype: C{str}
    @return: The result of rendering the fragment.
    """
    page = LivePage(docFactory=stan(fragment))
    fragment.setFragmentParent(page)
    (request, context) = _makeContext()
    page.renderHTTP(context)
    page.action_close(context)
    return request.v



def renderPlainFragment(fragment):
    """
    same as L{render}, but expects an L{nevow.rend.Fragment} or any
    other L{nevow.inevow.IRenderer}
    """
    page = Page(docFactory=stan(fragment))
    (request, context) = _makeContext()
    page.renderHTTP(context)
    return request.v




"""
Test for upgrading a WebSite to move its TCP and SSL information onto separate
objects.
"""

from twisted.application.service import IService
from twisted.cred.portal import IRealm

from nevow.inevow import IResource

from axiom.test.historic.stubloader import StubbedTest
from axiom.dependency import installedOn
from axiom.userbase import LoginSystem

from xmantissa.port import TCPPort, SSLPort
from xmantissa.web import SiteConfiguration
from xmantissa.website import WebSite
from xmantissa.publicweb import AnonymousSite
from xmantissa.ixmantissa import IMantissaSite, IWebViewer

from xmantissa.test.historic.stub_website4to5 import cert


class WebSiteUpgradeTests(StubbedTest):
    expectedHostname = u"example.net"

    def test_preservedAttributes(self):
        """
        Test that some data from the simple parts of the schema is preserved.
        """
        site = self.store.findUnique(SiteConfiguration)
        self.assertEqual(site.httpLog, self.store.filesdir.child('httpd.log'))
        self.assertEqual(site.hostname, self.expectedHostname)


    def test_portNumber(self):
        """
        Test that the WebSite's portNumber attribute is transformed into a
        TCPPort instance.
        """
        site = self.store.findUnique(SiteConfiguration)
        ports = list(self.store.query(TCPPort, TCPPort.factory == site))
        self.assertEqual(len(ports), 1)
        self.assertEqual(ports[0].portNumber, 8088)
        self.assertEqual(installedOn(ports[0]), self.store)
        self.assertEqual(list(self.store.interfacesFor(ports[0])), [IService])


    def test_securePortNumber(self):
        """
        Test that the WebSite's securePortNumber attribute is transformed into
        an SSLPort instance.
        """
        site = self.store.findUnique(SiteConfiguration)
        ports = list(self.store.query(SSLPort, SSLPort.factory == site))
        self.assertEqual(len(ports), 1)
        self.assertEqual(ports[0].portNumber, 6443)
        certPath = self.store.newFilePath('server.pem')
        self.assertEqual(ports[0].certificatePath, certPath)
        self.assertEqual(certPath.getContent(), cert)
        self.assertEqual(installedOn(ports[0]), self.store)
        self.assertEqual(list(self.store.interfacesFor(ports[0])), [IService])


    def test_deleted(self):
        """
        The L{WebSite} should no longer exist in the site store.
        """
        self.assertEqual(list(self.store.query(WebSite)), [])


    def test_anonymousSite(self):
        """
        An L{AnonymousSite} is created and installed on the site store.
        """
        resource = self.store.findUnique(AnonymousSite)
        self.assertEqual(list(self.store.interfacesFor(resource)),
                         [IResource, IMantissaSite, IWebViewer])
        self.assertIdentical(installedOn(resource), self.store)
        self.assertIdentical(resource.loginSystem, IRealm(self.store))


    def test_singleLoginSystem(self):
        """
        The upgrade should not create extra L{LoginSystem} items.
        """
        self.assertEqual(self.store.query(LoginSystem).count(), 1)


    def test_userStore(self):
        """
        Test that WebSites in user stores upgrade without errors.
        """
        ls = self.store.findUnique(LoginSystem)
        substore = ls.accountByAddress(u'testuser', u'localhost').avatars.open()
        d = substore.whenFullyUpgraded()
        def fullyUpgraded(ignored):
            web = substore.findUnique(WebSite)
            self.assertEqual(web.hitCount, 321)
        return d.addCallback(fullyUpgraded)


    def tearDown(self):
        d = StubbedTest.tearDown(self)
        def flushit(ign):
            from epsilon.cooperator import SchedulerStopped
            self.flushLoggedErrors(SchedulerStopped)
            return ign
        return d.addCallback(flushit)



from axiom.test.historic.stubloader import saveStub

from xmantissa.signup import UserInfoSignup, Multifactor
from xmantissa.webadmin import AdministrativeBenefactor

def createDatabase(s):
    ab = AdministrativeBenefactor(store=s)
    mf = Multifactor(store=s)
    mf.add(ab)
    UserInfoSignup(store=s,
                   prefixURL=u'/a/b',
                   booth=s,
                   benefactor=mf,
                   emailTemplate=u'TEMPLATE!',
                   prompt=u'OK?')

if __name__ == '__main__':
    saveStub(createDatabase, 10664)



from axiom.test.historic.stubloader import saveStub
from xmantissa.signup import PasswordReset

def createDatabase(s):
    PasswordReset(store=s, installedOn=s)

if __name__ == '__main__':
    saveStub(createDatabase, 8991)




from axiom.test.historic.stubloader import saveStub

from xmantissa.stats import StatBucket

def createDatabase(s):
    StatBucket(store=s,
               type=u"_axiom_query:select 1")
    StatBucket(store=s,
               type=u"axiom_commits")

if __name__ == '__main__':
    saveStub(createDatabase, 7071)



# -*- test-case-name: xmantissa.test.test_people -*-
# Copyright 2008 Divmod, Inc. See LICENSE file for details

"""
Person item and related functionality.
"""

from warnings import warn

try:
    from PIL import Image
except ImportError:
    Image = None

# Python < 2.5 compatibility
try:
    from email.utils import getaddresses
except ImportError:
    from email.Utils import getaddresses

from zope.interface import implements

from twisted.python import components
from twisted.python.filepath import FilePath
from twisted.python.reflect import qual

from nevow import rend, athena, inevow, static, tags, url
from nevow.athena import expose, LiveElement
from nevow.loaders import stan
from nevow.page import Element, renderer
from nevow.taglibrary import tabbedPane
from formless import nameToLabel

from epsilon import extime
from epsilon.structlike import record
from epsilon.descriptor import requiredAttribute

from axiom import item, attributes
from axiom.tags import Tag
from axiom.dependency import dependsOn
from axiom.attributes import boolean
from axiom.upgrade import (
    registerUpgrader, registerAttributeCopyingUpgrader,
    registerDeletionUpgrader)
from axiom.userbase import LoginAccount, LoginMethod

from xmantissa.ixmantissa import IPeopleFilter
from xmantissa import ixmantissa, webnav, webtheme, liveform, signup
from xmantissa.ixmantissa import IOrganizerPlugin, IContactType
from xmantissa.webapp import PrivateApplication
from xmantissa.tdbview import TabularDataView, ColumnViewBase
from xmantissa.scrolltable import ScrollingElement, UnsortableColumn
from xmantissa.fragmentutils import dictFillSlots
from xmantissa.webtheme import ThemedDocumentFactory


def makeThumbnail(inputFile, outputFile, thumbnailSize, outputFormat='jpeg'):
    """
    Make a thumbnail of the image stored at C{inputPath}, preserving its
    aspect ratio, and write the result to C{outputPath}.

    @param inputFile: The image file (or path to the file) to thumbnail.
    @type inputFile: C{file} or C{str}

    @param outputFile: The file (or path to the file) to write the thumbnail
    to.
    @type outputFile: C{file} or C{str}

    @param thumbnailSize: The maximum length (in pixels) of the longest side of
    the thumbnail image.
    @type thumbnailSize: C{int}

    @param outputFormat: The C{format} argument to pass to L{Image.save}.
    Defaults to I{jpeg}.
    @type format: C{str}
    """
    if Image is None:
        # throw the ImportError here
        import PIL
    image = Image.open(inputFile)
    # Resize needed?
    if thumbnailSize < max(image.size):
        # Convert bilevel and paletted images to grayscale and RGB respectively;
        # otherwise PIL silently switches to Image.NEAREST sampling.
        if image.mode == '1':
            image = image.convert('L')
        elif image.mode == 'P':
            image = image.convert('RGB')
        image.thumbnail((thumbnailSize, thumbnailSize), Image.ANTIALIAS)
    image.save(outputFile, outputFormat)



def _normalizeWhitespace(text):
    """
    Remove leading and trailing whitespace and collapse adjacent spaces into a
    single space.

    @type text: C{unicode}
    @rtype: C{unicode}
    """
    return u' '.join(text.split())



def _objectToName(o):
    """
    Derive a possibly-useful string type name from C{o}'s class.

    @rtype: C{str}
    """
    return nameToLabel(o.__class__.__name__).lstrip()



def _descriptiveIdentifier(contactType):
    """
    Get a descriptive identifier for C{contactType}, taking into account the
    fact that it might not have implemented the C{descriptiveIdentifier}
    method.

    @type contactType: L{IContactType} provider.

    @rtype: C{unicode}
    """
    descriptiveIdentifierMethod = getattr(
        contactType, 'descriptiveIdentifier', None)
    if descriptiveIdentifierMethod is not None:
        return descriptiveIdentifierMethod()
    warn(
        "IContactType now has the 'descriptiveIdentifier'"
        " method, %s did not implement it" % (contactType.__class__,),
        category=PendingDeprecationWarning)
    return _objectToName(contactType).decode('ascii')



def _organizerPluginName(plugin):
    """
    Get a name for C{plugin}, taking into account the fact that it might not
    have defined L{IOrganizerPlugin.name}.

    @type plugin: L{IOrganizerPlugin} provider.

    @rtype: C{unicode}
    """
    name = getattr(plugin, 'name', None)
    if name is not None:
        return name
    warn(
        "IOrganizerPlugin now has the 'name' attribute"
        " and %s does not define it" % (plugin.__class__,),
        category=PendingDeprecationWarning)
    return _objectToName(plugin).decode('ascii')



class ContactGroup(record('groupName')):
    """
    An object describing a group of L{IContactItem} implementors.

    @see IContactType.getContactGroup

    @ivar groupName: The name of the group.
    @type groupNode: C{unicode}
    """



class BaseContactType(object):
    """
    Base class for L{IContactType} implementations which provides useful
    default behavior.
    """
    allowMultipleContactItems = True

    def uniqueIdentifier(self):
        """
        Uniquely identify this contact type.
        """
        return qual(self.__class__).decode('ascii')


    def getParameters(self, contact):
        """
        Return a list of L{liveform.Parameter} objects to be used to create
        L{liveform.LiveForm}s suitable for creating or editing contact
        information of this type.

        Override this in a subclass.

        @param contact: A contact item, values from which should be used as
            defaults in the parameters.  C{None} if the parameters are for
            creating a new contact item.

        """
        raise NotImplementedError("%s did not implement getParameters" % (self,))


    def coerce(self, **kw):
        """
        Callback for input validation.

        @param **kw: Mapping of submitted parameter names to values.

        @rtype: C{dict}
        @return: Mapping of coerced parameter names to values.
        """
        return kw


    def getEditFormForPerson(self, person):
        """
        Return C{None}.
        """
        return None


    def getContactGroup(self, contactItem):
        """
        Return C{None}.
        """
        return None



class SimpleReadOnlyView(Element):
    """
    Simple read-only contact item view, suitable for returning from an
    implementation of L{IContactType.getReadOnlyView}.

    @ivar attribute: The contact item attribute which should be rendered.
    @type attribute: Axiom attribute (e.g. L{attributes.text})

    @ivar contactItem: The contact item.
    @type contactItem: L{item.Item}
    """
    docFactory = ThemedDocumentFactory(
        'person-contact-read-only-view', 'store')

    def __init__(self, attribute, contactItem):
        Element.__init__(self)
        self.attribute = attribute
        self.contactItem = contactItem
        self.store = contactItem.store


    def attributeName(self, req, tag):
        """
        Render the name of L{contactItem}'s class, e.g. "Email Address".
        """
        return nameToLabel(self.contactItem.__class__.__name__)
    renderer(attributeName)


    def attributeValue(self, req, tag):
        """
        Render the value of L{attribute} on L{contactItem}.
        """
        return self.attribute.__get__(self.contactItem)
    renderer(attributeValue)



class _PersonVIPStatus:
    """
    Contact item type used by L{VIPPersonContactType}.

    @param person: The person whose VIP status we're interested in.
    @type person: L{Person}
    """
    def __init__(self, person):
        self.person = person



class VIPPersonContactType(BaseContactType):
    """
    A contact type for controlling whether L{Person.vip} is set.
    """
    implements(IContactType)
    allowMultipleContactItems = False

    def getParameters(self, contactItem):
        """
        Return a list containing a single parameter suitable for changing the
        VIP status of a person.

        @type contactItem: L{_PersonVIPStatus}

        @rtype: C{list} of L{liveform.Parameter}
        """
        isVIP = False # default
        if contactItem is not None:
            isVIP = contactItem.person.vip
        return [liveform.Parameter(
            'vip', liveform.CHECKBOX_INPUT, bool, 'VIP', default=isVIP)]


    def getContactItems(self, person):
        """
        Return a list containing a L{_PersonVIPStatus} instance for C{person}.

        @type person: L{Person}

        @rtype: C{list} of L{_PersonVIPStatus}
        """
        return [_PersonVIPStatus(person)]


    def createContactItem(self, person, vip):
        """
        Set the VIP status of C{person} to C{vip}.

        @type person: L{Person}

        @type vip: C{bool}

        @rtype: L{_PersonVIPStatus}
        """
        person.vip = vip
        return _PersonVIPStatus(person)


    def editContactItem(self, contactItem, vip):
        """
        Change the VIP status of C{contactItem}'s person to C{vip}.

        @type contactItem: L{_PersonVIPStatus}

        @type vip: C{bool}

        @rtype: C{NoneType}
        """
        contactItem.person.vip = vip


    def getReadOnlyView(self, contactItem):
        """
        Return a fragment which will render as the empty string.
        L{PersonSummaryView} handles the rendering of VIP status in the
        read-only L{Person} view.

        @type contactItem: L{_PersonVIPStatus}

        @rtype: L{Element}
        """
        return Element(docFactory=stan(tags.invisible()))



class EmailContactType(BaseContactType):
    """
    Contact type plugin which allows a person to have an email address.

    @ivar store: The L{Store} the contact items will be created in.
    """
    implements(IContactType)

    def __init__(self, store):
        self.store = store


    def getParameters(self, emailAddress):
        """
        Return a C{list} of one L{LiveForm} parameter for editing an
        L{EmailAddress}.

        @type emailAddress: L{EmailAddress} or C{NoneType}
        @param emailAddress: If not C{None}, an existing contact item from
            which to get the email address default value.

        @rtype: C{list}
        @return: The parameters necessary for specifying an email address.
        """
        if emailAddress is not None:
            address = emailAddress.address
        else:
            address = u''
        return [
            liveform.Parameter('email', liveform.TEXT_INPUT,
                               _normalizeWhitespace, 'Email Address',
                               default=address)]


    def descriptiveIdentifier(self):
        """
        Return 'Email Address'
        """
        return u'Email Address'


    def _existing(self, email):
        """
        Return the existing L{EmailAddress} item with the given address, or
        C{None} if there isn't one.
        """
        return self.store.findUnique(
            EmailAddress,
            EmailAddress.address == email,
            default=None)


    def createContactItem(self, person, email):
        """
        Create a new L{EmailAddress} associated with the given person based on
        the given email address.

        @type person: L{Person}
        @param person: The person with whom to associate the new
            L{EmailAddress}.

        @type email: C{unicode}
        @param email: The value to use for the I{address} attribute of the
            newly created L{EmailAddress}.  If C{''}, no L{EmailAddress} will
            be created.

        @return: C{None}
        """
        if email:
            address = self._existing(email)
            if address is not None:
                raise ValueError('There is already a person with that email '
                                 'address (%s): ' % (address.person.name,))
            return EmailAddress(store=self.store,
                                address=email,
                                person=person)


    def getContactItems(self, person):
        """
        Return all L{EmailAddress} instances associated with the given person.

        @type person: L{Person}
        """
        return person.store.query(
            EmailAddress,
            EmailAddress.person == person)


    def editContactItem(self, contact, email):
        """
        Change the email address of the given L{EmailAddress} to that specified
        by C{email}.

        @type email: C{unicode}
        @param email: The new value to use for the I{address} attribute of the
            L{EmailAddress}.

        @return: C{None}
        """
        address = self._existing(email)
        if address is not None and address is not contact:
            raise ValueError('There is already a person with that email '
                             'address (%s): ' % (address.person.name,))
        contact.address = email


    def getReadOnlyView(self, contact):
        """
        Return a L{SimpleReadOnlyView} for the given L{EmailAddress}.
        """
        return SimpleReadOnlyView(EmailAddress.address, contact)



class PeopleBenefactor(item.Item):
    implements(ixmantissa.IBenefactor)
    endowed = attributes.integer(default=0)
    powerupNames = ["xmantissa.people.AddPerson"]



class Person(item.Item):
    """
    Person Per"son (p[~e]r"s'n; 277), n.

        1. A character or part, as in a play; a specific kind or manifestation
        of individual character, whether in real life, or in literary or
        dramatic representation; an assumed character. [Archaic] [1913 Webster]

    This is Mantissa's simulation of a person, which has attached contact
    information.  It is highly pluggable, mostly via the L{Organizer} object.

    Do not create this item directly, as functionality of L{IOrganizerPlugin}
    powerups will be broken if you do.  Instead, use L{Organizer.createPerson}.
    """

    typeName = 'mantissa_person'
    schemaVersion = 3

    organizer = attributes.reference(
        doc="""
        The L{Organizer} to which this Person belongs.
        """)

    name = attributes.text(
        doc="""
        This name of this person.
        """, caseSensitive=False)

    created = attributes.timestamp(defaultFactory=extime.Time)

    vip = boolean(
        doc="""
        Flag indicating this L{Person} is very important.
        """, default=False, allowNone=False)


    def getDisplayName(self):
        return self.name


    def getEmailAddresses(self):
        """
        Return an iterator of all email addresses associated with this person.

        @return: an iterator of unicode strings in RFC2822 address format.
        """
        return self.store.query(
            EmailAddress,
            EmailAddress.person == self).getColumn('address')

    def getEmailAddress(self):
        """
        Return the default email address associated with this person.

        Note: this is effectively random right now if a person has more than
        one address.  It's just the first address returned.  This should be
        fixed in a future version.

        @return: a unicode string in RFC2822 address format.
        """
        for a in self.getEmailAddresses():
            return a


    def getMugshot(self):
        """
        Return the L{Mugshot} associated with this L{Person}, or an unstored
        L{Mugshot} pointing at a placeholder mugshot image.
        """
        mugshot = self.store.findUnique(
            Mugshot, Mugshot.person == self, default=None)
        if mugshot is not None:
            return mugshot
        return Mugshot.placeholderForPerson(self)


    def registerExtract(self, extract, timestamp=None):
        """
        @param extract: some Item that implements L{inevow.IRenderer}
        """
        if timestamp is None:
            timestamp = extime.Time()

        return ExtractWrapper(store=self.store,
                              extract=extract,
                              timestamp=timestamp,
                              person=self)


    def getExtractWrappers(self, n):
        return self.store.query(ExtractWrapper,
                                ExtractWrapper.person == self,
                                sort=ExtractWrapper.timestamp.desc,
                                limit=n)


item.declareLegacyItem(
    Person.typeName,
    1,
    dict(organizer=attributes.reference(),
         name=attributes.text(caseSensitive=True),
         created=attributes.timestamp()))

registerAttributeCopyingUpgrader(Person, 1, 2)

item.declareLegacyItem(
    Person.typeName,
    2,
    dict(organizer=attributes.reference(),
         name=attributes.text(caseSensitive=True),
         created=attributes.timestamp(),
         vip=attributes.boolean(default=False, allowNone=False)))

registerAttributeCopyingUpgrader(Person, 2, 3)



class ExtractWrapper(item.Item):
    extract = attributes.reference(whenDeleted=attributes.reference.CASCADE)
    timestamp = attributes.timestamp(indexed=True)
    person = attributes.reference(reftype=Person,
                                  whenDeleted=attributes.reference.CASCADE)



def _stringifyKeys(d):
    """
    Return a copy of C{d} with C{str} keys.

    @type d: C{dict} with C{unicode} keys.
    @rtype: C{dict} with C{str} keys.
    """
    return dict((k.encode('ascii'), v)  for (k, v) in d.iteritems())



class AllPeopleFilter(object):
    """
    L{IPeopleFilter} which includes all L{Person} items from the given store
    in its query.
    """
    implements(IPeopleFilter)
    filterName = 'All'

    def getPeopleQueryComparison(self, store):
        """
        @see IPeopleFilter.getPeopleQueryComparison
        """
        return None



class VIPPeopleFilter(object):
    """
    L{IPeopleFilter} which includes all VIP L{Person} items from the given
    store in its query.
    """
    implements(IPeopleFilter)
    filterName = 'VIP'

    def getPeopleQueryComparison(self, store):
        """
        @see IPeopleFilter.getPeopleQueryComparison
        """
        return Person.vip == True



class TaggedPeopleFilter(record('filterName')):
    """
    L{IPeopleFilter} which includes in its query all L{Person} items to which
    a specific tag has been applied.
    """
    implements(IPeopleFilter)

    def getPeopleQueryComparison(self, store):
        """
        @see IPeopleFilter.getPeopleQueryComparison
        """
        return attributes.AND(
                Tag.object == Person.storeID,
                Tag.name == self.filterName)



class BaseOrganizerPlugin(object):
    """
    Base class for L{IOrganizerPlugin} implementations, which provides null
    implementations of the interface's callback/notification methods.
    """
    name = requiredAttribute('name')

    def personCreated(self, person):
        """
        Do nothing.

        @see IOrganizerPlugin.personCreated
        """


    def personNameChanged(self, person, name):
        """
        Do nothing.

        @see IOrganizerPlugin.personNameChanged
        """


    def contactItemCreated(self, item):
        """
        Do nothing.

        @see IOrganizerPlugin.contactItemCreated
        """


    def contactItemEdited(self, item):
        """
        Do nothing.

        @see IOrganizerPlugin.contactItemEdited
        """



class ContactInfoOrganizerPlugin(BaseOrganizerPlugin):
    """
    Trivial in-memory L{IOrganizerPlugin}.
    """
    implements(IOrganizerPlugin)

    name = u'Contact'

    def getContactTypes(self):
        """
        No contact types.

        @see IOrganizerPlugin.getContactTypes
        """
        return ()


    def getPeopleFilters(self):
        """
        No people filters.

        @see IOrganizerPlugin.getPeopleFilters
        """
        return ()


    def personalize(self, person):
        """
        Return a L{ReadOnlyContactInfoView} for C{person}.

        @see IOrganizerPlugin.personalize
        """
        return ReadOnlyContactInfoView(person)



class Organizer(item.Item):
    """
    Oversee the creation, location, destruction, and modification of
    people in a particular set (eg, the set of people you know).
    """
    implements(ixmantissa.INavigableElement)

    typeName = 'mantissa_people'
    schemaVersion = 3

    _webTranslator = dependsOn(PrivateApplication)

    storeOwnerPerson = attributes.reference(
        doc="A L{Person} representing the owner of the store this organizer lives in",
        reftype=Person,
        whenDeleted=attributes.reference.DISALLOW)


    powerupInterfaces = (ixmantissa.INavigableElement,)


    def __init__(self, *a, **k):
        super(Organizer, self).__init__(*a, **k)
        if 'storeOwnerPerson' not in k:
            self.storeOwnerPerson = self._makeStoreOwnerPerson()


    def _makeStoreOwnerPerson(self):
        """
        Make a L{Person} representing the owner of the store that this
        L{Organizer} is installed in.

        @rtype: L{Person}
        """
        if self.store is None:
            return None
        userInfo = self.store.findFirst(signup.UserInfo)
        name = u''
        if userInfo is not None:
            name = userInfo.realName
        account = self.store.findUnique(LoginAccount,
                                        LoginAccount.avatars == self.store, None)
        ownerPerson = self.createPerson(name)
        if account is not None:
            for method in (self.store.query(
                    LoginMethod,
                    attributes.AND(LoginMethod.account == account,
                                   LoginMethod.internal == False))):
                self.createContactItem(
                    EmailContactType(self.store),
                    ownerPerson, dict(
                        email=method.localpart + u'@' + method.domain))
        return ownerPerson


    def getOrganizerPlugins(self):
        """
        Return an iterator of the installed L{IOrganizerPlugin} powerups.
        """
        return (list(self.store.powerupsFor(IOrganizerPlugin))
            + [ContactInfoOrganizerPlugin()])


    def _gatherPluginMethods(self, methodName):
        """
        Walk through each L{IOrganizerPlugin} powerup, yielding the bound
        method if the powerup implements C{methodName}.  Upon encountering a
        plugin which fails to implement it, issue a
        L{PendingDeprecationWarning}.

        @param methodName: The name of a L{IOrganizerPlugin} method.
        @type methodName: C{str}

        @return: Iterable of methods.
        """
        for plugin in self.getOrganizerPlugins():
            implementation = getattr(plugin, methodName, None)
            if implementation is not None:
                yield implementation
            else:
                warn(
                    ('IOrganizerPlugin now has the %r method, %s'
                        ' did not implement it') % (
                            methodName, plugin.__class__),
                    category=PendingDeprecationWarning)


    def _checkContactType(self, contactType):
        """
        Possibly emit some warnings about C{contactType}'s implementation of
        L{IContactType}.

        @type contactType: L{IContactType} provider
        """
        if getattr(contactType, 'getEditFormForPerson', None) is None:
            warn(
                "IContactType now has the 'getEditFormForPerson'"
                " method, but %s did not implement it." % (
                    contactType.__class__,),
                category=PendingDeprecationWarning)

        if getattr(contactType, 'getEditorialForm', None) is not None:
            warn(
                "The IContactType %s defines the 'getEditorialForm'"
                " method, which is deprecated.  'getEditFormForPerson'"
                " does something vaguely similar." % (contactType.__class__,),
                category=DeprecationWarning)


    def getContactTypes(self):
        """
        Return an iterator of L{IContactType} providers available to this
        organizer's store.
        """
        yield VIPPersonContactType()
        yield EmailContactType(self.store)
        yield PostalContactType()
        yield PhoneNumberContactType()
        yield NotesContactType()
        for getContactTypes in self._gatherPluginMethods('getContactTypes'):
            for contactType in getContactTypes():
                self._checkContactType(contactType)
                yield contactType


    def getPeopleFilters(self):
        """
        Return an iterator of L{IPeopleFilter} providers available to this
        organizer's store.
        """
        yield AllPeopleFilter()
        yield VIPPeopleFilter()
        for getPeopleFilters in self._gatherPluginMethods('getPeopleFilters'):
            for peopleFilter in getPeopleFilters():
                yield peopleFilter
        for tag in sorted(self.getPeopleTags()):
            yield TaggedPeopleFilter(tag)


    def getPeopleTags(self):
        """
        Return a sequence of tags which have been applied to L{Person} items.

        @rtype: C{set}
        """
        query = self.store.query(
            Tag, Tag.object == Person.storeID)
        return set(query.getColumn('name').distinct())


    def groupReadOnlyViews(self, person):
        """
        Collect all contact items from the available contact types for the
        given person, organize them by contact group, and turn them into
        read-only views.

        @type person: L{Person}
        @param person: The person whose contact items we're interested in.

        @return: A mapping of of L{ContactGroup} names to the read-only views
        of their member contact items, with C{None} being the key for
        groupless contact items.
        @rtype: C{dict} of C{str}
        """
        # this is a slightly awkward, specific API, but at the time of
        # writing, read-only views are the thing that the only caller cares
        # about.  we need the contact type to get a read-only view for a
        # contact item.  there is no way to get from a contact item to a
        # contact type, so this method can't be "groupContactItems" (which
        # seems to make more sense), unless it returned some weird data
        # structure which managed to associate contact items and contact
        # types.
        grouped = {}
        for contactType in self.getContactTypes():
            for contactItem in contactType.getContactItems(person):
                contactGroup = contactType.getContactGroup(contactItem)
                if contactGroup is not None:
                    contactGroup = contactGroup.groupName
                if contactGroup not in grouped:
                    grouped[contactGroup] = []
                grouped[contactGroup].append(
                    contactType.getReadOnlyView(contactItem))
        return grouped


    def getContactCreationParameters(self):
        """
        Yield a L{Parameter} for each L{IContactType} known.

        Each yielded object can be used with a L{LiveForm} to create a new
        instance of a particular L{IContactType}.
        """
        for contactType in self.getContactTypes():
            if contactType.allowMultipleContactItems:
                descriptiveIdentifier = _descriptiveIdentifier(contactType)
                yield liveform.ListChangeParameter(
                    contactType.uniqueIdentifier(),
                    contactType.getParameters(None),
                    defaults=[],
                    modelObjects=[],
                    modelObjectDescription=descriptiveIdentifier)
            else:
                yield liveform.FormParameter(
                    contactType.uniqueIdentifier(),
                    liveform.LiveForm(
                        lambda **k: k,
                        contactType.getParameters(None)))


    def _parametersToDefaults(self, parameters):
        """
        Extract the defaults from C{parameters}, constructing a dictionary
        mapping parameter names to default values, suitable for passing to
        L{ListChangeParameter}.

        @type parameters: C{list} of L{liveform.Parameter} or
        L{liveform.ChoiceParameter}.

        @rtype: C{dict}
        """
        defaults = {}
        for p in parameters:
            if isinstance(p, liveform.ChoiceParameter):
                selected = []
                for choice in p.choices:
                    if choice.selected:
                        selected.append(choice.value)
                defaults[p.name] = selected
            else:
                defaults[p.name] = p.default
        return defaults


    def toContactEditorialParameter(self, contactType, person):
        """
        Convert the given contact type into a L{liveform.LiveForm} parameter.

        @type contactType: L{IContactType} provider.

        @type person: L{Person}

        @rtype: L{liveform.Parameter} or similar.
        """
        contactItems = list(contactType.getContactItems(person))
        if contactType.allowMultipleContactItems:
            defaults = []
            modelObjects = []
            for contactItem in contactItems:
                defaultedParameters = contactType.getParameters(contactItem)
                if defaultedParameters is None:
                    continue
                defaults.append(self._parametersToDefaults(
                    defaultedParameters))
                modelObjects.append(contactItem)
            descriptiveIdentifier = _descriptiveIdentifier(contactType)
            return liveform.ListChangeParameter(
                contactType.uniqueIdentifier(),
                contactType.getParameters(None),
                defaults=defaults,
                modelObjects=modelObjects,
                modelObjectDescription=descriptiveIdentifier)
        (contactItem,) = contactItems
        return liveform.FormParameter(
            contactType.uniqueIdentifier(),
            liveform.LiveForm(
                lambda **k: k,
                contactType.getParameters(contactItem)))


    def getContactEditorialParameters(self, person):
        """
        Yield L{LiveForm} parameters to edit each contact item of each contact
        type for the given person.

        @type person: L{Person}
        @return: An iterable of two-tuples.  The first element of each tuple
            is an L{IContactType} provider.  The third element of each tuple
            is the L{LiveForm} parameter object for that contact item.
        """
        for contactType in self.getContactTypes():
            yield (
                contactType,
                self.toContactEditorialParameter(contactType, person))


    _NO_VIP = object()

    def createPerson(self, nickname, vip=_NO_VIP):
        """
        Create a new L{Person} with the given name in this organizer.

        @type nickname: C{unicode}
        @param nickname: The value for the new person's C{name} attribute.

        @type vip: C{bool}
        @param vip: Value to set the created person's C{vip} attribute to
        (deprecated).

        @rtype: L{Person}
        """
        for person in (self.store.query(
                Person, attributes.AND(
                    Person.name == nickname,
                    Person.organizer == self))):
            raise ValueError("Person with name %r exists already." % (nickname,))
        person = Person(
            store=self.store,
            created=extime.Time(),
            organizer=self,
            name=nickname)

        if vip is not self._NO_VIP:
            warn(
                "Usage of Organizer.createPerson's 'vip' parameter"
                " is deprecated",
                category=DeprecationWarning)
            person.vip = vip

        self._callOnOrganizerPlugins('personCreated', person)
        return person


    def createContactItem(self, contactType, person, contactInfo):
        """
        Create a new contact item for the given person with the given contact
        type.  Broadcast a creation to all L{IOrganizerPlugin} powerups.

        @type contactType: L{IContactType}
        @param contactType: The contact type which will be used to create the
            contact item.

        @type person: L{Person}
        @param person: The person with whom the contact item will be
            associated.

        @type contactInfo: C{dict}
        @param contactInfo: The contact information to use to create the
            contact item.

        @return: The contact item, as created by the given contact type.
        """
        contactItem = contactType.createContactItem(
            person, **_stringifyKeys(contactInfo))
        if contactItem is not None:
            self._callOnOrganizerPlugins('contactItemCreated', contactItem)
        return contactItem


    def editContactItem(self, contactType, contactItem, contactInfo):
        """
        Edit the given contact item with the given contact type.  Broadcast
        the edit to all L{IOrganizerPlugin} powerups.

        @type contactType: L{IContactType}
        @param contactType: The contact type which will be used to edit the
            contact item.

        @param contactItem: The contact item to edit.

        @type contactInfo: C{dict}
        @param contactInfo: The contact information to use to edit the
            contact item.

        @return: C{None}
        """
        contactType.editContactItem(
            contactItem, **_stringifyKeys(contactInfo))
        self._callOnOrganizerPlugins('contactItemEdited', contactItem)


    def _callOnOrganizerPlugins(self, methodName, *args):
        """
        Call a method on all L{IOrganizerPlugin} powerups on C{self.store}, or
        emit a deprecation warning for each one which does not implement that
        method.
        """
        for observer in self.getOrganizerPlugins():
            method = getattr(observer, methodName, None)
            if method is not None:
                method(*args)
            else:
                warn(
                    "IOrganizerPlugin now has the %s method, %s "
                    "did not implement it" % (methodName, observer.__class__,),
                    category=PendingDeprecationWarning)


    def editPerson(self, person, nickname, edits):
        """
        Change the name and contact information associated with the given
        L{Person}.

        @type person: L{Person}
        @param person: The person which will be modified.

        @type nickname: C{unicode}
        @param nickname: The new value for L{Person.name}

        @type edits: C{list}
        @param edits: list of tuples of L{IContactType} providers and
        corresponding L{ListChanges} objects or dictionaries of parameter
        values.
        """
        for existing in self.store.query(Person, Person.name == nickname):
            if existing is person:
                continue
            raise ValueError(
                "A person with the name %r exists already." % (nickname,))
        oldname = person.name
        person.name = nickname
        self._callOnOrganizerPlugins('personNameChanged', person, oldname)
        for contactType, submission in edits:
            if contactType.allowMultipleContactItems:
                for edit in submission.edit:
                    self.editContactItem(
                        contactType, edit.object, edit.values)
                for create in submission.create:
                    create.setter(
                        self.createContactItem(
                            contactType, person, create.values))
                for delete in submission.delete:
                    delete.deleteFromStore()
            else:
                (contactItem,) = contactType.getContactItems(person)
                self.editContactItem(
                    contactType, contactItem, submission)


    def deletePerson(self, person):
        """
        Delete the given person from the store.
        """
        person.deleteFromStore()


    def personByName(self, name):
        """
        Retrieve the L{Person} item for the given Q2Q address,
        creating it first if necessary.

        @type name: C{unicode}
        """
        return self.store.findOrCreate(Person, organizer=self, name=name)


    def personByEmailAddress(self, address):
        """
        Retrieve the L{Person} item for the given email address
        (or return None if no such person exists)

        @type name: C{unicode}
        """
        email = self.store.findUnique(EmailAddress,
                                      EmailAddress.address == address,
                                      default=None)
        if email is not None:
            return email.person


    def linkToPerson(self, person):
        """
        @param person: L{Person} instance
        @return: string url at which C{person} will be rendered
        """
        return self._webTranslator.linkTo(person.storeID)


    def urlForViewState(self, person, viewState):
        """
        Return a url for L{OrganizerFragment} which will display C{person} in
        state C{viewState}.

        @type person: L{Person}
        @type viewState: L{ORGANIZER_VIEW_STATES} constant.

        @rtype: L{url.URL}
        """
        # ideally there would be a more general mechanism for encoding state
        # like this in a url, rather than ad-hoc query arguments for each
        # fragment which needs to do it.
        organizerURL = self._webTranslator.linkTo(self.storeID)
        return url.URL(
            netloc='', scheme='',
            pathsegs=organizerURL.split('/')[1:],
            querysegs=(('initial-person', person.name),
                       ('initial-state', viewState)))


    # INavigableElement
    def getTabs(self):
        """
        Implement L{INavigableElement.getTabs} to return a single tab,
        'People', that points to this item.
        """
        return [webnav.Tab('People', self.storeID, 0.5, authoritative=True)]



def organizer1to2(old):
    o = old.upgradeVersion(old.typeName, 1, 2)
    o._webTranslator = old.store.findOrCreate(PrivateApplication)
    return o

registerUpgrader(organizer1to2, Organizer.typeName, 1, 2)



item.declareLegacyItem(Organizer.typeName, 2,
    dict(_webTranslator=attributes.reference()))



registerAttributeCopyingUpgrader(Organizer, 2, 3)



class VIPColumn(UnsortableColumn):
    def getType(self):
        return 'boolean'



class MugshotURLColumn(record('organizer attributeID')):
    """
    L{ixmantissa.IColumn} provider which extracts the URL of a L{Person}'s
    mugshot.

    @type organizer: L{Organizer}
    """
    implements(ixmantissa.IColumn)

    def extractValue(self, model, item):
        """
        Figure out the URL of C{item}'s mugshot.

        @type item: L{Person}

        @rtype: C{unicode}
        """
        return self.organizer.linkToPerson(item) + u'/mugshot/smaller'


    def sortAttribute(self):
        """
        Return C{None}.  Unsortable.
        """
        return None


    def getType(self):
        """
        Return C{text}
        """
        return 'text'


    def toComparableValue(self, value):
        """
        This shouldn't be called.

        @raise L{NotImplementedError}: Always.
        """
        raise NotImplementedError(
            '%r does not implement toComparableValue: it is unsortable' %
            (self.__class__,))



class PersonScrollingFragment(ScrollingElement):
    """
    Scrolling element which displays L{Person} objects and allows actions to
    be taken on them.

    @type organizer: L{Organizer}
    """
    jsClass = u'Mantissa.People.PersonScroller'

    def __init__(self, organizer, baseConstraint, defaultSortColumn,
            webTranslator):
        self._originalBaseConstraint = baseConstraint
        ScrollingElement.__init__(
            self,
            organizer.store,
            Person,
            baseConstraint,
            [MugshotURLColumn(organizer, 'mugshotURL'),
             VIPColumn(Person.vip, 'vip'),
             Person.name],
            defaultSortColumn=defaultSortColumn,
            webTranslator=webTranslator)
        self.organizer = organizer
        self.filters = dict((filter.filterName, filter)
                for filter in organizer.getPeopleFilters())


    def getInitialArguments(self):
        """
        Include L{organizer}'s C{storeOwnerPerson}'s name.
        """
        return (ScrollingElement.getInitialArguments(self)
                    + [self.organizer.storeOwnerPerson.name])


    def filterByFilter(self, filterName):
        """
        Swap L{baseConstraint} with the result of calling
        L{IPeopleFilter.getPeopleQueryComparison} on the named filter.

        @type filterName: C{unicode}
        """
        filter = self.filters[filterName]
        self.baseConstraint = filter.getPeopleQueryComparison(self.store)
    expose(filterByFilter)



class PersonSummaryView(Element):
    """
    Fragment which renders a business card-like summary of a L{Person}: their
    mugshot, vip status, and name.

    @type person: L{Person}
    @ivar person: The person to summarize.
    """
    docFactory = ThemedDocumentFactory('person-summary', 'store')

    def __init__(self, person):
        self.person = person
        self.organizer = person.organizer
        self.store = person.store


    def mugshotURL(self, req, tag):
        """
        Render the URL of L{person}'s mugshot, or the URL of a placeholder
        mugshot if they don't have one set.
        """
        return self.organizer.linkToPerson(self.person) + '/mugshot/smaller'
    renderer(mugshotURL)


    def personName(self, req, tag):
        """
        Render the display name of L{person}.
        """
        return self.person.getDisplayName()
    renderer(personName)


    def vipStatus(self, req, tag):
        """
        Return C{tag} if L{person} is a VIP, otherwise return the empty
        string.
        """
        if self.person.vip:
            return tag
        return ''
    renderer(vipStatus)



class ReadOnlyContactInfoView(Element):
    """
    Fragment which renders a read-only version of a person's contact
    information.

    @ivar person: A person.
    @type person: L{Person}
    """
    docFactory = ThemedDocumentFactory(
        'person-read-only-contact-info', 'store')

    def __init__(self, person):
        self.person = person
        self.organizer = person.organizer
        self.store = person.store
        Element.__init__(self)


    def personSummary(self, request, tag):
        """
        Render a L{PersonSummaryView} for L{person}.
        """
        return PersonSummaryView(self.person)
    renderer(personSummary)


    def contactInfo(self, request, tag):
        """
        Render the result of calling L{IContactType.getReadOnlyView} on the
        corresponding L{IContactType} for each piece of contact info
        associated with L{person}.  Arrange the result by group, using
        C{tag}'s I{contact-group} pattern. Groupless contact items will have
        their views yielded directly.

        The I{contact-group} pattern appears once for each distinct
        L{ContactGroup}, with the following slots filled:
          I{name} - The group's C{groupName}.
          I{views} - A sequence of read-only views belonging to the group.
        """
        groupPattern = inevow.IQ(tag).patternGenerator('contact-group')
        groupedViews = self.organizer.groupReadOnlyViews(self.person)
        for (groupName, views) in groupedViews.iteritems():
            if groupName is None:
                yield views
            else:
                yield groupPattern().fillSlots(
                    'name', groupName).fillSlots(
                    'views', views)
    renderer(contactInfo)



class ORGANIZER_VIEW_STATES:
    """
    Some constants describing possible initial states of L{OrganizerFragment}.

    @ivar EDIT: The state which involves editing a person.

    @ivar ALL_STATES: A sequence of all valid initial states.
    """
    EDIT = u'edit'

    ALL_STATES = (EDIT,)



class _ElementWrapper(LiveElement):
    """
    L{LiveElement} which wraps & renders an L{Element}.

    @type wrapped: L{Element}
    """
    docFactory = stan(
        tags.div(render=tags.directive('liveElement'))[
            tags.directive('element')])

    def __init__(self, wrapped):
        LiveElement.__init__(self)
        self.wrapped = wrapped


    def element(self, request, tag):
        """
        Render L{wrapped}.
        """
        return self.wrapped
    renderer(element)



class PersonPluginView(LiveElement):
    """
    Element which renders UI for selecting between the views of available
    person plugins.  A tab will be rendered for each L{IOrganizerPlugin} in
    L{plugins}, with the corresponding personalization being rendered when a
    tab is selected.

    @ivar plugins: Sequence of L{IOrganizerPlugin} providers.
    @type plugins: C{list}

    @ivar person: The person we're interested in.
    @type person: L{Person}
    """
    docFactory = ThemedDocumentFactory('person-plugins', 'store')
    jsClass = u'Mantissa.People.PersonPluginView'

    def __init__(self, plugins, person):
        LiveElement.__init__(self)
        self.plugins = plugins
        self.person = person
        self.store = person.store


    def pluginTabbedPane(self, request, tag):
        """
        Render a L{tabbedPane.TabbedPaneFragment} with an entry for each item
        in L{plugins}.
        """
        iq = inevow.IQ(tag)
        tabNames = [
            _organizerPluginName(p).encode('ascii') # gunk
                for p in self.plugins]
        child = tabbedPane.TabbedPaneFragment(
            zip(tabNames,
                ([self.getPluginWidget(tabNames[0])]
                    + [iq.onePattern('pane-body') for _ in tabNames[1:]])))
        child.jsClass = u'Mantissa.People.PluginTabbedPane'
        child.setFragmentParent(self)
        return child
    renderer(pluginTabbedPane)


    def _toLiveElement(self, element):
        """
        Wrap the given element in a L{LiveElement} if it is not already one.

        @rtype: L{LiveElement}
        """
        if isinstance(element, LiveElement):
            return element
        return _ElementWrapper(element)


    def getPluginWidget(self, pluginName):
        """
        Return the named plugin's view.

        @type pluginName: C{unicode}
        @param pluginName: The name of the plugin.

        @rtype: L{LiveElement}
        """
        # this will always pick the first plugin with pluginName if there is
        # more than one.  don't do that.
        for plugin in self.plugins:
            if _organizerPluginName(plugin) == pluginName:
                view = self._toLiveElement(
                    plugin.personalize(self.person))
                view.setFragmentParent(self)
                return view
    expose(getPluginWidget)



class OrganizerFragment(LiveElement):
    """
    Address book view.  The initial state of this fragment can be extracted
    from the query parameters in its url, if present.  The two parameters it
    looks for are: I{initial-person} (the name of the L{Person} to select
    initially in the scrolltable) and I{initial-state} (a
    L{ORGANIZER_VIEW_STATES} constant describing what to do with the person).
    Both query arguments must be present if either is.

    @type organizer: L{Organizer}
    @ivar organizer: The organizer for which this is a view.

    @ivar initialPerson: The person to load initially.  Defaults to C{None}.
    @type initialPerson: L{Person} or C{NoneType}

    @ivar initialState: The initial state of the organizer view.  Defaults to
    C{None}.
    @type initialState: L{ORGANIZER_VIEW_STATES} or C{NoneType}
    """
    docFactory = ThemedDocumentFactory('people-organizer', 'store')
    fragmentName = None
    live = 'athena'
    title = 'People'
    jsClass = u'Mantissa.People.Organizer'

    def __init__(self, organizer, initialPerson=None, initialState=None):
        LiveElement.__init__(self)
        self.organizer = organizer
        self.initialPerson = initialPerson
        self.initialState = initialState

        self.store = organizer.store
        self.wt = organizer._webTranslator


    def head(self):
        """
        Do nothing.
        """


    def beforeRender(self, ctx):
        """
        Implement this hook to initialize the L{initialPerson} and
        L{initialState} slots with information from the request url's query
        args.
        """
        # see the comment in Organizer.urlForViewState which suggests an
        # alternate implementation of this kind of functionality.
        request = inevow.IRequest(ctx)
        if not set(['initial-person', 'initial-state']).issubset( # <=
            set(request.args)):
            return
        initialPersonName = request.args['initial-person'][0].decode('utf-8')
        initialPerson = self.store.findFirst(
            Person, Person.name == initialPersonName)
        if initialPerson is None:
            return
        initialState = request.args['initial-state'][0].decode('utf-8')
        if initialState not in ORGANIZER_VIEW_STATES.ALL_STATES:
            return
        self.initialPerson = initialPerson
        self.initialState = initialState


    def getInitialArguments(self):
        """
        Include L{organizer}'s C{storeOwnerPerson}'s name, and the name of
        L{initialPerson} and the value of L{initialState}, if they are set.
        """
        initialArguments = (self.organizer.storeOwnerPerson.name,)
        if self.initialPerson is not None:
            initialArguments += (self.initialPerson.name, self.initialState)
        return initialArguments


    def getAddPerson(self):
        """
        Return an L{AddPersonFragment} which is a child of this fragment and
        which will add a person to C{self.organizer}.
        """
        fragment = AddPersonFragment(self.organizer)
        fragment.setFragmentParent(self)
        return fragment
    expose(getAddPerson)


    def getImportPeople(self):
        """
        Return an L{ImportPeopleWidget} which is a child of this fragment and
        which will add people to C{self.organizer}.
        """
        fragment = ImportPeopleWidget(self.organizer)
        fragment.setFragmentParent(self)
        return fragment
    expose(getImportPeople)


    def getEditPerson(self, name):
        """
        Get an L{EditPersonView} for editing the person named C{name}.

        @param name: A person name.
        @type name: C{unicode}

        @rtype: L{EditPersonView}
        """
        view = EditPersonView(self.organizer.personByName(name))
        view.setFragmentParent(self)
        return view
    expose(getEditPerson)


    def deletePerson(self, name):
        """
        Delete the person named C{name}

        @param name: A person name.
        @type name: C{unicode}
        """
        self.organizer.deletePerson(self.organizer.personByName(name))
    expose(deletePerson)


    def peopleTable(self, request, tag):
        """
        Return a L{PersonScrollingFragment} which will display the L{Person}
        items in the wrapped organizer.
        """
        f = PersonScrollingFragment(
            self.organizer, None, Person.name, self.wt)
        f.setFragmentParent(self)
        f.docFactory = webtheme.getLoader(f.fragmentName)
        return f
    renderer(peopleTable)


    def peopleFilters(self, request, tag):
        """
        Return an instance of C{tag}'s I{filter} pattern for each filter we
        get from L{Organizer.getPeopleFilters}, filling the I{name} slot with
        the filter's name.  The first filter will be rendered using the
        I{selected-filter} pattern.
        """
        filters = iter(self.organizer.getPeopleFilters())
        # at some point we might actually want to look at what filter is
        # yielded first, and filter the person list accordingly.  we're just
        # going to assume it's the "All" filter, and leave the person list
        # untouched for now.
        yield tag.onePattern('selected-filter').fillSlots(
            'name', filters.next().filterName)
        pattern = tag.patternGenerator('filter')
        for filter in filters:
            yield pattern.fillSlots('name', filter.filterName)
    renderer(peopleFilters)


    def getPersonPluginWidget(self, name):
        """
        Return the L{PersonPluginView} for the named person.

        @type name: C{unicode}
        @param name: A value which corresponds to the I{name} attribute of an
        extant L{Person}.

        @rtype: L{PersonPluginView}
        """
        fragment = PersonPluginView(
            self.organizer.getOrganizerPlugins(),
            self.organizer.personByName(name))
        fragment.setFragmentParent(self)
        return fragment
    expose(getPersonPluginWidget)

components.registerAdapter(OrganizerFragment, Organizer, ixmantissa.INavigableFragment)



class EditPersonView(LiveElement):
    """
    Render a view for editing the contact information for a L{Person}.

    @ivar person: L{Person} which can be edited.

    @ivar contactTypes: A mapping from parameter names to the L{IContactTypes}
        whose items the parameters are editing.
    """
    docFactory = ThemedDocumentFactory('edit-person', 'store')
    fragmentName = 'edit-person'
    jsClass = u'Mantissa.People.EditPerson'

    def __init__(self, person):
        athena.LiveElement.__init__(self)
        self.person = person
        self.store = person.store
        self.organizer = person.organizer
        self.contactTypes = {}


    def editContactItems(self, nickname, **edits):
        """
        Update the information on the contact items associated with the wrapped
        L{Person}.

        @type nickname: C{unicode}
        @param nickname: New value to use for the I{name} attribute of the
            L{Person}.

        @param **edits: mapping from contact type identifiers to
            ListChanges instances.
        """
        submissions = []
        for paramName, submission in edits.iteritems():
            contactType = self.contactTypes[paramName]
            submissions.append((contactType, submission))
        self.person.store.transact(
            self.organizer.editPerson,
            self.person, nickname, submissions)


    def makeEditorialLiveForms(self):
        """
        Make some L{liveform.LiveForm} instances for editing the contact
        information of the wrapped L{Person}.
        """
        parameters = [
            liveform.Parameter(
                'nickname', liveform.TEXT_INPUT,
                _normalizeWhitespace, 'Name',
                default=self.person.name)]
        separateForms = []
        for contactType in self.organizer.getContactTypes():
            if getattr(contactType, 'getEditFormForPerson', None):
                editForm = contactType.getEditFormForPerson(self.person)
                if editForm is not None:
                    editForm.setFragmentParent(self)
                    separateForms.append(editForm)
                    continue
            param = self.organizer.toContactEditorialParameter(
                contactType, self.person)
            parameters.append(param)
            self.contactTypes[param.name] = contactType
        form = liveform.LiveForm(
            self.editContactItems, parameters, u'Save')
        form.compact()
        form.jsClass = u'Mantissa.People.EditPersonForm'
        form.setFragmentParent(self)
        return [form] + separateForms


    def mugshotFormURL(self, request, tag):
        """
        Render a URL for L{MugshotUploadForm}.
        """
        return self.organizer.linkToPerson(self.person) + '/mugshotUploadForm'
    renderer(mugshotFormURL)


    def editorialContactForms(self, request, tag):
        """
        Put the result of L{makeEditorialLiveForms} in C{tag}.
        """
        return tag[self.makeEditorialLiveForms()]
    renderer(editorialContactForms)



class RealName(item.Item):
    """
    This is a legacy item left over from a previous schema.  Do not create it.
    """
    typeName = 'mantissa_organizer_addressbook_realname'
    schemaVersion = 2

    empty = attributes.reference()


item.declareLegacyItem(
    RealName.typeName, 1,
    dict(person=attributes.reference(
            doc="""
            allowNone=False,
            whenDeleted=attributes.reference.CASCADE,
            reftype=Person
            """),
         first=attributes.text(),
         last=attributes.text(indexed=True)))

registerDeletionUpgrader(RealName, 1, 2)



class EmailAddress(item.Item):
    """
    An email address contact info item associated with a L{Person}.

    Do not create this item directly, as functionality of L{IOrganizerPlugin}
    powerups will be broken if you do.  Instead, use
    L{Organizer.createContactItem} with L{EmailContactType}.
    """
    typeName = 'mantissa_organizer_addressbook_emailaddress'
    schemaVersion = 3

    address = attributes.text(allowNone=False)
    person = attributes.reference(
        allowNone=False,
        whenDeleted=attributes.reference.CASCADE,
        reftype=Person)
    label = attributes.text(
        """
        This is a label for the role of the email address, usually something like
        "home", "work", "school".
        """,
        allowNone=False,
        default=u'')

def emailAddress1to2(old):
    return old.upgradeVersion('mantissa_organizer_addressbook_emailaddress',
                              1, 2,
                              address=old.address,
                              person=old.person)

registerUpgrader(emailAddress1to2,
                 'mantissa_organizer_addressbook_emailaddress',
                 1, 2)

item.declareLegacyItem(EmailAddress.typeName, 2, dict(
                  address = attributes.text(allowNone=False),
                  person = attributes.reference(allowNone=False)))

registerAttributeCopyingUpgrader(EmailAddress, 2, 3)


class PhoneNumber(item.Item):
    """
    A contact information item representing a L{Person}'s phone number.

    Do not create this item directly, as functionality of L{IOrganizerPlugin}
    powerups will be broken if you do.  Instead, use
    L{Organizer.createContactItem} with L{PhoneNumberContactType}.
    """
    typeName = 'mantissa_organizer_addressbook_phonenumber'
    schemaVersion = 3

    number = attributes.text(allowNone=False)
    person = attributes.reference(allowNone=False)
    label = attributes.text(
        """
        This is a label for the role of the phone number.
        """,
        allowNone=False,
        default=u'',)


    class LABELS:
        """
        Constants to use for the value of the L{label} attribute, describing
        the type of the telephone number.

        @ivar HOME: This is a home phone number.
        @type HOME: C{unicode}

        @ivar WORK: This is a work phone number.
        @type WORK: C{unicode}

        @ivar MOBILE: This is a mobile phone number.
        @type MOBILE: C{unicode}

        @ivar HOME_FAX: This is the 80's and someone has a fax machine in
        their house.
        @type HOME_FAX: C{unicode}

        @ivar WORK_FAX: This is the 80's and someone has a fax machine in
        their office.
        @type WORK_FAX: C{unicode}

        @ivar PAGER: This is the 80's and L{person} is a drug dealer.
        @type PAGER: C{unicode}

        @ivar ALL_LABELS: A sequence of all of the label constants.
        @type ALL_LABELS: C{list}
        """
        HOME = u'Home'
        WORK = u'Work'
        MOBILE = u'Mobile'
        HOME_FAX = u'Home Fax'
        WORK_FAX = u'Work Fax'
        PAGER = u'Pager'


        ALL_LABELS = [HOME, WORK, MOBILE, HOME_FAX, WORK_FAX, PAGER]

def phoneNumber1to2(old):
    return old.upgradeVersion('mantissa_organizer_addressbook_phonenumber',
                              1, 2,
                              number=old.number,
                              person=old.person)

item.declareLegacyItem(PhoneNumber.typeName, 2, dict(
                  number = attributes.text(allowNone=False),
                  person = attributes.reference(allowNone=False)))

registerUpgrader(phoneNumber1to2,
                 'mantissa_organizer_addressbook_phonenumber',
                 1, 2)

registerAttributeCopyingUpgrader(PhoneNumber, 2, 3)



class PhoneNumberContactType(BaseContactType):
    """
    Contact type plugin which allows telephone numbers to be associated with
    people.
    """
    implements(IContactType)

    def getParameters(self, phoneNumber):
        """
        Return a C{list} of two liveform parameters, one for editing
        C{phoneNumber}'s I{number} attribute, and one for editing its I{label}
        attribute.

        @type phoneNumber: L{PhoneNumber} or C{NoneType}
        @param phoneNumber: If not C{None}, an existing contact item from
        which to get the parameter's default values.

        @rtype: C{list}
        """
        defaultNumber = u''
        defaultLabel = PhoneNumber.LABELS.HOME
        if phoneNumber is not None:
            defaultNumber = phoneNumber.number
            defaultLabel = phoneNumber.label
        labelChoiceParameter = liveform.ChoiceParameter(
            'label',
            [liveform.Option(label, label, label == defaultLabel)
                for label in PhoneNumber.LABELS.ALL_LABELS],
            'Number Type')
        return [
            labelChoiceParameter,
            liveform.Parameter(
                'number',
                liveform.TEXT_INPUT,
                unicode,
                'Phone Number',
                default=defaultNumber)]


    def descriptiveIdentifier(self):
        """
        Return 'Phone Number'
        """
        return u'Phone Number'


    def createContactItem(self, person, label, number):
        """
        Create a L{PhoneNumber} item for C{number}, associated with C{person}.

        @type person: L{Person}

        @param label: The value to use for the I{label} attribute of the new
        L{PhoneNumber} item.
        @type label: C{unicode}

        @param number: The value to use for the I{number} attribute of the new
        L{PhoneNumber} item.  If C{''}, no item will be created.
        @type number: C{unicode}

        @rtype: L{PhoneNumber} or C{NoneType}
        """
        if number:
            return PhoneNumber(
                store=person.store, person=person, label=label, number=number)


    def editContactItem(self, contact, label, number):
        """
        Change the I{number} attribute of C{contact} to C{number}, and the
        I{label} attribute to C{label}.

        @type contact: L{PhoneNumber}

        @type label: C{unicode}

        @type number: C{unicode}

        @return: C{None}
        """
        contact.label = label
        contact.number = number


    def getContactItems(self, person):
        """
        Return an iterable of L{PhoneNumber} items that are associated with
        C{person}.

        @type person: L{Person}
        """
        return person.store.query(
            PhoneNumber, PhoneNumber.person == person)


    def getReadOnlyView(self, contact):
        """
        Return a L{ReadOnlyPhoneNumberView} for the given L{PhoneNumber}.
        """
        return ReadOnlyPhoneNumberView(contact)



class ReadOnlyPhoneNumberView(Element):
    """
    Read-only view for L{PhoneNumber}.

    @type phoneNumber: L{PhoneNumber}
    """
    docFactory = ThemedDocumentFactory(
        'person-contact-read-only-phone-number-view', 'store')

    def __init__(self, phoneNumber):
        self.phoneNumber = phoneNumber
        self.store = phoneNumber.store


    def number(self, request, tag):
        """
        Add the value of L{phoneNumber}'s I{number} attribute to C{tag}.
        """
        return tag[self.phoneNumber.number]
    renderer(number)


    def label(self, request, tag):
        """
        Add the value of L{phoneNumber}'s I{label} attribute to C{tag}.
        """
        return tag[self.phoneNumber.label]
    renderer(label)



class PostalAddress(item.Item):
    typeName = 'mantissa_organizer_addressbook_postaladdress'

    address = attributes.text(allowNone=False)
    person = attributes.reference(
        allowNone=False,
        whenDeleted=attributes.reference.CASCADE,
        reftype=Person)



class PostalContactType(BaseContactType):
    """
    Contact type plugin which allows a person to have a snail-mail address.
    """
    implements(IContactType)

    def getParameters(self, postalAddress):
        """
        Return a C{list} of one L{LiveForm} parameter for editing a
        L{PostalAddress}.

        @type postalAddress: L{PostalAddress} or C{NoneType}

        @param postalAddress: If not C{None}, an existing contact item from
            which to get the postal address default value.

        @rtype: C{list}
        @return: The parameters necessary for specifying a postal address.
        """
        address = u''
        if postalAddress is not None:
            address = postalAddress.address
        return [
            liveform.Parameter('address', liveform.TEXT_INPUT,
                               unicode, 'Postal Address', default=address)]


    def descriptiveIdentifier(self):
        """
        Return 'Postal Address'
        """
        return u'Postal Address'


    def createContactItem(self, person, address):
        """
        Create a new L{PostalAddress} associated with the given person based on
        the given postal address.

        @type person: L{Person}
        @param person: The person with whom to associate the new
            L{EmailAddress}.

        @type address: C{unicode}
        @param address: The value to use for the I{address} attribute of the
            newly created L{PostalAddress}.  If C{''}, no L{PostalAddress} will
            be created.

        @rtype: L{PostalAddress} or C{NoneType}
        """
        if address:
            return PostalAddress(
                store=person.store, person=person, address=address)


    def editContactItem(self, contact, address):
        """
        Change the postal address of the given L{PostalAddress} to that
        specified by C{address}.

        @type contact: L{PostalAddress}
        @param contact: The existing contact item to modify.

        @type address: C{unicode}
        @param address: The new value to use for the I{address} attribute of
            the L{PostalAddress}.

        @return: C{None}
        """
        contact.address = address


    def getContactItems(self, person):
        """
        Return a C{list} of the L{PostalAddress} items associated with the
        given person.

        @type person: L{Person}
        """
        return person.store.query(PostalAddress, PostalAddress.person == person)


    def getReadOnlyView(self, contact):
        """
        Return a L{SimpleReadOnlyView} for the given L{PostalAddress}.
        """
        return SimpleReadOnlyView(PostalAddress.address, contact)



class Notes(item.Item):
    typeName = 'mantissa_organizer_addressbook_notes'

    notes = attributes.text(allowNone=False)
    person = attributes.reference(allowNone=False)



class NotesContactType(BaseContactType):
    """
    Contact type plugin which allows a person to be annotated with a free-form
    textual note.
    """
    implements(IContactType)
    allowMultipleContactItems = False

    def getParameters(self, notes):
        """
        Return a C{list} of one L{LiveForm} parameter for editing a
        L{Notes}.

        @type notes: L{Notes} or C{NoneType}
        @param notes: If not C{None}, an existing contact item from
            which to get the parameter's default value.

        @rtype: C{list}
        """
        defaultNotes = u''
        if notes is not None:
            defaultNotes = notes.notes
        return [
            liveform.Parameter('notes', liveform.TEXTAREA_INPUT,
                               unicode, 'Notes', default=defaultNotes)]


    def descriptiveIdentifier(self):
        """
        Return 'Notes'
        """
        return u'Notes'


    def createContactItem(self, person, notes):
        """
        Create a new L{Notes} associated with the given person based on the
        given string.

        @type person: L{Person}
        @param person: The person with whom to associate the new L{Notes}.

        @type notes: C{unicode}
        @param notes: The value to use for the I{notes} attribute of the newly
        created L{Notes}.  If C{''}, no L{Notes} will be created.

        @rtype: L{Notes} or C{NoneType}
        """
        if notes:
            return Notes(
                store=person.store, person=person, notes=notes)


    def editContactItem(self, contact, notes):
        """
        Set the I{notes} attribute of C{contact} to the value of the C{notes}
        parameter.

        @type contact: L{Notes}
        @param contact: The existing contact item to modify.

        @type notes: C{unicode}
        @param notes: The new value to use for the I{notes} attribute of
            the L{Notes}.

        @return: C{None}
        """
        contact.notes = notes


    def getContactItems(self, person):
        """
        Return a C{list} of the L{Notes} items associated with the given
        person.  If none exist, create one, wrap it in a list and return it.

        @type person: L{Person}
        """
        notes = list(person.store.query(Notes, Notes.person == person))
        if not notes:
            return [Notes(store=person.store,
                          person=person,
                          notes=u'')]
        return notes


    def getReadOnlyView(self, contact):
        """
        Return a L{SimpleReadOnlyView} for the given L{Notes}.
        """
        return SimpleReadOnlyView(Notes.notes, contact)



class AddPerson(item.Item):
    implements(ixmantissa.INavigableElement)

    typeName = 'mantissa_add_person'
    schemaVersion = 2

    powerupInterfaces = (ixmantissa.INavigableElement,)
    organizer = dependsOn(Organizer)

    def getTabs(self):
        return []



def addPerson1to2(old):
    ap = old.upgradeVersion(old.typeName, 1, 2)
    ap.organizer = old.store.findOrCreate(Organizer)
    return ap

registerUpgrader(addPerson1to2, AddPerson.typeName, 1, 2)



class AddPersonFragment(athena.LiveFragment):
    """
    View class for L{AddPerson}, presenting a user interface for creating a new
    L{Person}.

    @ivar organizer: The L{Organizer} instance which will be used to add the
        person.
    """
    docFactory = ThemedDocumentFactory('add-person', 'store')
    jsClass = u'Mantissa.People.AddPerson'

    def __init__(self, organizer):
        athena.LiveFragment.__init__(self)
        self.organizer = organizer
        self.store = organizer.store


    def head(self):
        """
        Supply no content to the head area of the page.
        """
        return None


    _baseParameters = [
        liveform.Parameter('nickname', liveform.TEXT_INPUT,
                           _normalizeWhitespace, 'Name')]

    def render_addPersonForm(self, ctx, data):
        """
        Create and return a L{liveform.LiveForm} for creating a new L{Person}.
        """
        addPersonForm = liveform.LiveForm(
            self.addPerson, self._baseParameters, description='Add Person')
        addPersonForm.compact()
        addPersonForm.jsClass = u'Mantissa.People.AddPersonForm'
        addPersonForm.setFragmentParent(self)
        return addPersonForm


    def addPerson(self, nickname):
        """
        Create a new L{Person} with the given C{nickname}.

        @type nickname: C{unicode}
        @param nickname: The value for the I{name} attribute of the created
            L{Person}.

        @raise L{liveform.InputError}: When some aspect of person creation
        raises a L{ValueError}.
        """
        try:
            self.organizer.createPerson(nickname)
        except ValueError, e:
            raise liveform.InputError(unicode(e))
    expose(addPerson)



class ImportPeopleWidget(athena.LiveElement):
    """
    Widget that implements importing people to an L{Organizer}.

    @ivar organizer: the L{Organizer} to use
    """

    docFactory = ThemedDocumentFactory('import-people', 'store')

    jsClass = u'Mantissa.People.ImportPeopleWidget'

    def __init__(self, organizer):
        athena.LiveElement.__init__(self)
        self.organizer = organizer
        self.store = organizer.store


    def _parseAddresses(addresses):
        """
        Extract name/address pairs from an RFC 2822 style address list.

        For addresses without a display name, the name defaults to the
        local-part for the purpose of importing.

        @type addresses: unicode
        @return: a list of C{(name, email)} tuples
        """
        def coerce((name, email)):
            if len(email):
                if not len(name):
                    name = email.split(u'@', 1)[0]  # lame
                return (name, email)
        coerced = map(coerce, getaddresses([addresses]))
        return [r for r in coerced if r is not None]
    _parseAddresses = staticmethod(_parseAddresses)


    def importPeopleForm(self, request, tag):
        """
        Create and return a L{liveform.LiveForm} for adding new L{Person}s.
        """
        form = liveform.LiveForm(
            self.importAddresses,
            [liveform.Parameter('addresses', liveform.TEXTAREA_INPUT,
                                self._parseAddresses, 'Email Addresses')],
            description='Import People')
        form.jsClass = u'Mantissa.People.ImportPeopleForm'
        form.compact()
        form.setFragmentParent(self)
        return form
    importPeopleForm = renderer(importPeopleForm)


    def importAddresses(self, addresses):
        """
        Create new L{Person}s for the given names and email addresses.
        Names and emails that already exist are ignored.

        @param addresses: a sequence of C{(name, email)} tuples
                          (as returned from L{_parseAddresses})
        @return: the names of people actually imported
        """
        results = []
        for (name, address) in addresses:
            def txn():
                # Skip names and addresses that already exist.
                if self.organizer.personByEmailAddress(address) is not None:
                    return
                # XXX: Needs a better existence check.
                if self.store.query(Person, Person.name == name).count():
                    return

                try:
                    person = self.organizer.createPerson(name)
                    self.organizer.createContactItem(
                        EmailContactType(self.store), person,
                        dict(email=address))
                except ValueError, e:
                    # XXX: Granularity required;  see #711 and #2435
                    raise liveform.ConfigurationError(u'%r' % (e,))
                return person
            results.append(self.store.transact(txn))
        return [p.name for p in results if p is not None]



class PersonExtractFragment(TabularDataView):
    def render_navigation(self, ctx, data):
        return inevow.IQ(
                webtheme.getLoader('person-extracts')).onePattern('navigation')



class ExtractWrapperColumnView(ColumnViewBase):
    def stanFromValue(self, idx, item, value):
        return inevow.IRenderer(item.extract)



class MugshotUploadForm(rend.Page):
    """
    Resource which presents some UI for associating a new mugshot with
    L{person}.

    @ivar person: The person whose mugshot is going to be changed.
    @type person: L{Person}

    @ivar cbGotImage: Function to call after a successful upload.  It will be
    passed the C{unicode} content-type of the uploaded image and a file
    containing the uploaded image.
    """
    docFactory = ThemedDocumentFactory('mugshot-upload-form', 'store')

    def __init__(self, person, cbGotMugshot):
        rend.Page.__init__(self)
        self.person = person
        self.organizer  = person.organizer
        self.store = person.store
        self.cbGotMugshot = cbGotMugshot


    def renderHTTP(self, ctx):
        """
        Extract the data from the I{uploaddata} field of the request and pass
        it to our callback.
        """
        req = inevow.IRequest(ctx)
        if req.method == 'POST':
            udata = req.fields['uploaddata']
            self.cbGotMugshot(udata.type.decode('ascii'), udata.file)
        return rend.Page.renderHTTP(self, ctx)


    def render_smallerMugshotURL(self, ctx, data):
        """
        Render the URL of a smaller version of L{person}'s mugshot.
        """
        return self.organizer.linkToPerson(self.person) + '/mugshot/smaller'



class Mugshot(item.Item):
    """
    An image that is associated with a person
    """
    schemaVersion = 3

    type = attributes.text(doc="""
    Content-type of image data
    """, allowNone=False)

    body = attributes.path(doc="""
    Path to image data
    """, allowNone=False)

    # at the moment we require an upgrader to change the size of either of the
    # mugshot thumbnails.  we might save ourselves some effort by generating
    # scaled versions on demand, and caching them.
    smallerBody = attributes.path(doc="""
    Path to smaller version of image data
    """, allowNone=False)

    person = attributes.reference(doc="""
    L{Person} this mugshot is of
    """, allowNone=False)

    size = 120
    smallerSize = 60

    def fromFile(cls, person, inputFile, format):
        """
        Create a L{Mugshot} item for C{person} out of the image data in
        C{inputFile}, or update C{person}'s existing L{Mugshot} item to
        reflect the new images.

        @param inputFile: An image of a person.
        @type inputFile: C{file}

        @param person: The person this mugshot is to be associated with.
        @type person: L{Person}

        @param format: The format of the data in C{inputFile}.
        @type format: C{unicode} (e.g. I{jpeg})

        @rtype: L{Mugshot}
        """
        body = cls.makeThumbnail(inputFile, person, format, smaller=False)
        inputFile.seek(0)
        smallerBody = cls.makeThumbnail(
            inputFile, person, format, smaller=True)

        ctype = u'image/' + format

        self = person.store.findUnique(
            cls, cls.person == person, default=None)
        if self is None:
            self = cls(store=person.store,
                       person=person,
                       type=ctype,
                       body=body,
                       smallerBody=smallerBody)
        else:
            self.body = body
            self.smallerBody = smallerBody
            self.type = ctype
        return self
    fromFile = classmethod(fromFile)


    def makeThumbnail(cls, inputFile, person, format, smaller):
        """
        Make a thumbnail of a mugshot image and store it on disk.

        @param inputFile: The image to thumbnail.
        @type inputFile: C{file}

        @param person: The person this mugshot thumbnail is associated with.
        @type person: L{Person}

        @param format: The format of the data in C{inputFile}.
        @type format: C{str} (e.g. I{jpeg})

        @param smaller: Thumbnails are available in two sizes.  if C{smaller}
        is C{True}, then the thumbnail will be in the smaller of the two
        sizes.
        @type smaller: C{bool}

        @return: path to the thumbnail.
        @rtype: L{twisted.python.filepath.FilePath}
        """
        dirsegs = ['mugshots', str(person.storeID)]
        if smaller:
            dirsegs.insert(1, 'smaller')
            size = cls.smallerSize
        else:
            size = cls.size
        atomicOutputFile = person.store.newFile(*dirsegs)
        makeThumbnail(inputFile, atomicOutputFile, size, format)
        atomicOutputFile.close()
        return atomicOutputFile.finalpath
    makeThumbnail = classmethod(makeThumbnail)


    def placeholderForPerson(cls, person):
        """
        Make an unstored, placeholder L{Mugshot} instance for the given
        person.

        @param person: A person without a L{Mugshot}.
        @type person: L{Person}

        @rtype: L{Mugshot}
        """
        imageDir = FilePath(__file__).parent().child(
            'static').child('images')
        return cls(
            type=u'image/png',
            body=imageDir.child('mugshot-placeholder.png'),
            smallerBody=imageDir.child(
                'mugshot-placeholder-smaller.png'),
            person=person)
    placeholderForPerson = classmethod(placeholderForPerson)

def mugshot1to2(old):
    """
    Upgrader for L{Mugshot} from version 1 to version 2, which sets the
    C{smallerBody} attribute to the path of a smaller mugshot image.
    """
    smallerBody = Mugshot.makeThumbnail(old.body.open(),
                                        old.person,
                                        old.type.split('/')[1],
                                        smaller=True)

    return old.upgradeVersion(Mugshot.typeName, 1, 2,
                              person=old.person,
                              type=old.type,
                              body=old.body,
                              smallerBody=smallerBody)

registerUpgrader(mugshot1to2, Mugshot.typeName, 1, 2)



item.declareLegacyItem(
    Mugshot.typeName,
    2,
    dict(person=attributes.reference(),
         type=attributes.text(),
         body=attributes.path(),
         smallerBody=attributes.path()))



def mugshot2to3(old):
    """
    Upgrader for L{Mugshot} from version 2 to version 3, which re-thumbnails
    the mugshot to take into account the new value of L{Mugshot.smallerSize}.
    """
    new = old.upgradeVersion(Mugshot.typeName, 2, 3,
                             person=old.person,
                             type=old.type,
                             body=old.body,
                             smallerBody=old.smallerBody)
    new.smallerBody = new.makeThumbnail(
        new.body.open(), new.person, new.type[len('image/'):], smaller=True)
    return new

registerUpgrader(mugshot2to3, Mugshot.typeName, 2, 3)



class MugshotResource(rend.Page):
    """
    Web accessible resource that serves Mugshot images. Serves a smaller
    mugshot if the final path segment is "smaller"
    """
    smaller = False

    def __init__(self, mugshot):
        """
        @param mugshot: L{Mugshot}
        """
        self.mugshot = mugshot
        rend.Page.__init__(self)


    def locateChild(self, ctx, segments):
        if segments == ('smaller',):
            self.smaller = True
            return (self, ())
        return rend.NotFound


    def renderHTTP(self, ctx):
        if self.smaller:
            path = self.mugshot.smallerBody
        else:
            path = self.mugshot.body

        return static.File(path.path, str(self.mugshot.type))



def getPersonURL(person):
    """
    Return the address the view for this Person is available at.
    """
    return person.organizer.linkToPerson(person)



class PersonDetailFragment(athena.LiveFragment, rend.ChildLookupMixin):
    """
    Renderer for detailed information about a L{Person}.
    """
    fragmentName = 'person-detail'
    live = 'athena'

    def __init__(self, person):
        athena.LiveFragment.__init__(self, person)
        self.person = person


    def head(self):
        return None


    def _gotMugshotFile(self, ctype, infile):
        (majortype, minortype) = ctype.split('/')
        if majortype == 'image':
            Mugshot.fromFile(self.person, infile, minortype)


    def child_mugshotUploadForm(self, ctx):
        """
        Return a L{MugshotUploadForm}, which will render some UI for
        associating a new mugshot with this person.
        """
        return MugshotUploadForm(self.person, self._gotMugshotFile)


    def child_mugshot(self, ctx):
        """
        Return a L{MugshotResource} displaying this L{Person}'s mugshot image.
        """
        return MugshotResource(self.person.getMugshot())



components.registerAdapter(PersonDetailFragment, Person, ixmantissa.INavigableFragment)

class PersonFragment(rend.Fragment):
    def __init__(self, person, contactMethod=None):
        rend.Fragment.__init__(self, person,
                               webtheme.getLoader('person-fragment'))
        self.person = person
        self.contactMethod = contactMethod

    def render_person(self, ctx, data):
        detailURL = self.person.organizer.linkToPerson(self.person)

        mugshot = self.person.store.findUnique(Mugshot,
                                               Mugshot.person == self.person,
                                               default=None)
        if mugshot is None:
            mugshotURL = '/Mantissa/images/mugshot-placeholder-smaller.png'
        else:
            mugshotURL = detailURL + '/mugshot/smaller'

        name = self.person.getDisplayName()
        return dictFillSlots(ctx.tag, {'name': name,
                                       'detail-url': detailURL,
                                       'contact-method': self.contactMethod or name,
                                       'mugshot-url': mugshotURL})



# Copyright 2008 Divmod, Inc. See LICENSE file for details
# -*- test-case-name: xmantissa.test.test_webapp.AuthenticatedWebViewerTests,xmantissa.test.test_publicweb.AnonymousWebViewerTests -*-
"""
A collection of fake versions of various objects used in tests.

There are a lot of classes here because many of them have model/view
interactions that are expressed through adapter registrations, so having
additional types is helpful.
"""

from zope.interface import implements

from twisted.python.components import registerAdapter

from epsilon.structlike import record

from nevow.athena import LiveElement, LiveFragment
from nevow.page import Element
from nevow import rend, loaders
from nevow.inevow import IResource

from xmantissa.ixmantissa import INavigableFragment


class FakeLoader(record('name')):
    """
    A fake Nevow loader object.
    """



class FakeTheme(record('themeName loaders')):
    """
    A placeholder for theme lookup.

    @ivar loaders: A dict of strings to loader objects.

    @ivar themeName: A name that describes this theme.
    """

    def getDocFactory(self, name, default=None):
        """
        @param name: A loader name.
        """
        return self.loaders.get(name, default)



class FakeModel(object):
    """
    A simple 'model' object that does nothing, for the purposes of adaptation.
    """



class ResourceViewForFakeModel(rend.Page):
    """
    Implementor of L{IResource} for L{FakeModel}.
    """

registerAdapter(ResourceViewForFakeModel, FakeModel, IResource)



class _HasModel(object):
    """
    A mixin that provides a 'model' attribute for Element subclasses.

    This is a simple hack that attempts to cooperatively invoke __init__ so
    that its numerous subclasses don't have to define a constructor.  If you
    want to use it you should read its implementation.
    """
    def __init__(self, model):
        """
        Set the model attribute and delegate to the other subclass.
        """
        self.model = model
        self.__class__.__bases__[1].__init__(self)



class FakeElementModel(object):
    """
    A simple 'model' object that does nothing, for the purposes of adaptation.
    """



class ElementViewForFakeModel(_HasModel, Element):
    """
    L{Element} implementor of L{INavigableFragment} for L{FakeElementModel}.
    """
    implements(INavigableFragment)
    docFactory = loaders.stan('')

registerAdapter(ElementViewForFakeModel, FakeElementModel, INavigableFragment)



class FakeElementModelWithTheme(object):
    """
    A simple 'model' object that does nothing, for the purposes of adaptation
    to L{ElementViewForFakeModelWithTheme}.
    """



class ElementViewForFakeModelWithTheme(_HasModel, Element):
    """
    L{Element} implementor of L{INavigableFragment} for L{FakeElementModel}.
    """
    implements(INavigableFragment)
    fragmentName = 'awesome_page'

registerAdapter(ElementViewForFakeModelWithTheme, FakeElementModelWithTheme,
                INavigableFragment)



class FakeElementModelWithDocFactory(record('loader')):
    """
    A simple 'model' object that does nothing, for the purposes of adaptation
    to L{ElementViewForFakeModelWithDocFactory}.

    @ivar loader: A loader object (to be used as the view's docFactory).
    """



class ElementViewForFakeModelWithDocFactory(_HasModel, Element):
    """
    L{Element} implementor of L{INavigableFragment} for L{FakeElementModel}.
    """
    implements(INavigableFragment)

    def __init__(self, original):
        """
        Set docFactory and proceed as usual.
        """
        _HasModel.__init__(self, original)
        self.docFactory = original.loader



registerAdapter(ElementViewForFakeModelWithDocFactory, FakeElementModelWithDocFactory,
                INavigableFragment)



class FakeElementModelWithThemeAndDocFactory(record('fragmentName loader')):
    """
    A simple 'model' object that does nothing, for the purposes of adaptation
    to L{ElementViewForFakeModelWithThemeAndDocFactory}.
    """



class ElementViewForFakeModelWithThemeAndDocFactory(_HasModel, Element):
    """
    L{Element} implementor of L{INavigableFragment} for L{FakeElementModel}.
    """
    implements(INavigableFragment)

    def __init__(self, original):
        """
        Set docFactory and proceed as usual.
        """
        _HasModel.__init__(self, original)
        self.docFactory = original.loader
        self.fragmentName = original.fragmentName


registerAdapter(ElementViewForFakeModelWithThemeAndDocFactory,
                FakeElementModelWithThemeAndDocFactory,
                INavigableFragment)




class FakeFragmentModel(object):
    """
    A simple 'model' object that does nothing, for the purposes of adaptation.
    """



class FragmentViewForFakeModel(rend.Fragment):
    """
    L{Fragment} implementor of L{INavigableFragment} for L{FakeFragmentModel}.
    """
    implements(INavigableFragment)
    docFactory = loaders.stan('')

registerAdapter(FragmentViewForFakeModel, FakeFragmentModel, INavigableFragment)



class FakeLiveElementModel(object):
    """
    A simple 'model' object that does nothing, for the purposes of adaptation.
    """



class LiveElementViewForFakeModel(_HasModel, LiveElement):
    """
    L{LiveElement} Implementor of L{INavigableFragment} for
    L{FakeLiveElementModel}.
    """
    implements(INavigableFragment)
    docFactory = loaders.stan('')

registerAdapter(LiveElementViewForFakeModel, FakeLiveElementModel,
                INavigableFragment)



class FakeLiveFragmentModel(object):
    """
    A simple 'model' object that does nothing, for the purposes of adaptation.
    """



class LiveFragmentViewForFakeModel(LiveFragment):
    """
    L{LiveFragment} Implementor of L{INavigableFragment} for
    L{FakeLiveFragmentModel}.
    """
    implements(INavigableFragment)
    docFactory = loaders.stan('')

    @classmethod
    def wrap(cls, model):
        """
        Wrap the given model in this class.  Implement this as a method in this
        file so that the warning filename will match up...
        """
        return cls(model)

registerAdapter(LiveFragmentViewForFakeModel.wrap, FakeLiveFragmentModel,
                INavigableFragment)

class FakeElementModelWithLocateChildView(object):
    """
    A simple 'model' object that does nothing, for the purposes of adaptation.
    """
    def __init__(self, children, beLive):
        """
        @param children: an iterable of children to be returned from the view's
        locateChild.
        """
        self.childs = iter(children) # implemented this way because we want to
                                     # see an error if locateChild is called
                                     # too many times; often this will be a
                                     # sequence of length 1
        self.beLive = beLive


    def __conform__(self, interface):
        """
        @param interface: IResource (for which there is no adapter) or
        INavigableFragment (for which there is one, depending on this model's
        liveness).
        """
        if interface is not INavigableFragment:
            return None
        if self.beLive:
            return LiveElementViewForModelWithLocateChild(self)
        else:
            return ElementViewForFakeModelWithLocateChild(self)



class _HasLocateChild(_HasModel):
    """
    Has a locateChild that delegates to its model.
    """
    implements(INavigableFragment)
    docFactory = loaders.stan('')

    def locateChild(self, ctx, segments):
        """
        Stub implementation that merely records whether it was called.
        """
        return self.model.childs.next()



class LiveElementViewForModelWithLocateChild(_HasLocateChild, LiveElement):
    """
    Live element with a locateChild.
    """



class ElementViewForFakeModelWithLocateChild(_HasLocateChild, Element):
    """
    Non-live element with a locateChild.
    """


class FakeCustomizableElementModel(object):
    """
    A simple 'model' object that does nothing, for the purposes of adaptation.
    """
    username = None
    def custom(self, username):
        """
        Record the username our view was customized with.
        """
        self.username = username



class ElementViewForFakeCustomizableElementModel(_HasModel, Element):
    """
    An L{Element} that delegates C{customizeFor} calls to its model.
    """
    implements(INavigableFragment)
    docFactory = loaders.stan('')

    def customizeFor(self, username):
        """
        Delegate to model.
        """
        self.model.custom(username)
        return self

registerAdapter(ElementViewForFakeCustomizableElementModel,
                FakeCustomizableElementModel,
                INavigableFragment)


class FakeElementModelWithHead(record('head')):
    """
    A simple 'model' object that does nothing, for the purposes of adaptation.
    """

    def _head(self):
        return self.head


class ElementViewForFakeModelWithHead(_HasModel, Element):
    """
    L{Element} implementor of L{INavigableFragment} for L{FakeElementModel}.
    """
    implements(INavigableFragment)
    docFactory = loaders.stan('')

    def head(self):
        return self.model._head()

registerAdapter(ElementViewForFakeModelWithHead, FakeElementModelWithHead,
                INavigableFragment)



from axiom.test.historic import stubloader
from xmantissa.people import Organizer, Person, RealName
from xmantissa.webapp import PrivateApplication

class OrganizerTest(stubloader.StubbedTest):
    """
    Test L{Organizer}'s 2->3 upgrader.
    """
    def test_storeOwnerPerson(self):
        """
        Test that L{Organizer.storeOwnerPerson} is set to a L{Person} with no
        L{RealName}, since there isn't enough information in the store to
        construct one.
        """
        o = self.store.findUnique(Organizer)
        self.assertIdentical(o.storeOwnerPerson, self.store.findUnique(Person))
        self.assertEqual(self.store.count(RealName), 0)


    def test_webTranslator(self):
        """
        Test that L{Organizer._webTranslator} is preserved across versions.
        """
        o = self.store.findUnique(Organizer)
        self.assertIdentical(o._webTranslator, self.store.findUnique(PrivateApplication))



from axiom.test.historic import stubloader
from xmantissa.publicweb import FrontPage

class FrontPageTest(stubloader.StubbedTest):
    """
    Upgrader test for L{xmantissa.publicweb.FrontPage}.
    """
    def testUpgrade(self):
        """
        All the attributes of L{xmantissa.publicweb.FrontPage} are
        present after upgrading.
        """
        fp = self.store.findUnique(FrontPage)
        self.assertEqual(fp.publicViews, 17)
        self.assertEqual(fp.privateViews, 42)
        self.assertEqual(fp.prefixURL, u'')
        self.assertEqual(fp.defaultApplication, None)

